import numpy as np
import pyarrow.parquet as pq
import torch
from torch import nn as nn
from torch.nn import functional as tf

from ptls.custom_layers import Squeeze
from ptls.nn.seq_step import LastStepEncoder
from ptls.nn.normalization import L2NormEncoder
from ptls.data_load.padded_batch import PaddedBatch


class PerTransHead(nn.Module):
    def __init__(self, input_size):
        super().__init__()
        self.cnn = torch.nn.Conv1d(input_size, 1, 1)

    def forward(self, x):
        seq_len = x.payload.shape[1]
        feature_vec = torch.transpose(x.payload, 1, 2)

        tx = self.cnn(feature_vec)
        x = tf.avg_pool1d(tx, seq_len)
        x = torch.transpose(x, 1, 2)

        return x.squeeze()


class PerTransTransf(nn.Module):
    def __init__(self, input_size, out_size):
        super().__init__()
        self.cnn = torch.nn.Conv1d(input_size, out_size, 1)

    def forward(self, x):
        feature_vec = torch.transpose(x.payload, 1, 2)

        tx = self.cnn(feature_vec)
        out = torch.transpose(tx, 1, 2)

        return PaddedBatch(out, x.seq_lens)


class ConcatLenEncoder(nn.Module):
    def forward(self, x: PaddedBatch):
        lens = x.seq_lens.unsqueeze(-1).float().to(x.payload.device)
        lens_normed = lens / 200

        h = x.payload[range(len(x.payload)), [l - 1 for l in x.seq_lens]]

        embeddings = torch.cat([h, lens_normed, -torch.log(lens_normed)], -1)
        return embeddings


class MeanStepEncoder(nn.Module):
    def forward(self, x: PaddedBatch):
        means = torch.stack([e[0:l].mean(dim=0) for e, l in zip(x.payload, x.seq_lens)])
        return means


class PayloadEncoder(nn.Module):
    def forward(self, x: PaddedBatch):
        return x.payload


class AllStepsHead(nn.Module):
    def __init__(self, head):
        super().__init__()
        self.head = head

    def forward(self, x: PaddedBatch):
        out = self.head(x.payload)
        return PaddedBatch(out, x.seq_lens)


class AllStepsMeanHead(nn.Module):
    def __init__(self, head):
        super().__init__()
        self.head = head

    def forward(self, x: PaddedBatch):
        out = self.head(x.payload)
        means = torch.tensor([e[0:l].mean() for e, l in zip(out, x.seq_lens)])
        return means


class FlattenHead(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: PaddedBatch):
        mask = torch.zeros(x.payload.shape, dtype=bool)
        for i, length in enumerate(x.seq_lens):
            mask[i, :length] = True

        return x.payload.flatten()[mask.flatten()]


def scoring_head(input_size, params):
    """Deprecated. Use `ptls.models.Head` instead.
    """

    if params.pred_all_states and params.explicit_lengths:
        raise AttributeError('Can not use `pred_all_states` and `explicit_lengths` together for now')

    layers = []

    if not params.pred_all_states:
        if params.explicit_lengths:
            e0 = ConcatLenEncoder()
            input_size += 2
            layers.append(e0)
        else:
            e0 = LastStepEncoder()
            layers.append(e0)

    if params.norm_input:
        layers.append(L2NormEncoder())

    if params.use_batch_norm:
        layers.append(nn.BatchNorm1d(input_size))

    if params.get('neural_automl', False):
        if params.pred_all_states:
            raise AttributeError('neural_automl not supported with `pred_all_states` for now')
        layers.append(nn.BatchNorm1d(input_size))
        head = node.get_model(model_config=params.neural_automl, input_size=input_size)
        layers.append(head)
        return nn.Sequential(*layers)

    if "head_layers" not in params:
        head_output_size = params.get('num_classes', 1)
        if params.get('objective', 'classification') == 'classification':
            if head_output_size == 1:
                h = nn.Sequential(nn.Linear(input_size, head_output_size), nn.Sigmoid(), Squeeze())
            else:
                h = nn.Sequential(nn.Linear(input_size, head_output_size), nn.LogSoftmax(dim=1))
        else:
            h = nn.Sequential(nn.Linear(input_size, head_output_size), Squeeze())
        if params.pred_all_states:
            if params.pred_all_states_mean:
                h = AllStepsMeanHead(h)
                layers.append(h)
            else:
                h = AllStepsHead(h)
                layers.append(h)
                if params.get('pred_flatten', False):
                    layers.append(FlattenHead())

        else:
            layers.append(h)
    else:
        head_layers = [input_size] + params.head_layers + [1]
        for size_in, size_out in zip(head_layers[:-1], head_layers[1:]):
            layers.append(nn.Linear(size_in, size_out))
            if size_out != 1:
                layers.append(nn.BatchNorm1d(size_out))
                layers.append(nn.ReLU())
                layers.append(nn.Dropout(0.1))
            else:
                layers.append(nn.Sigmoid())
        layers.append(Squeeze())

    return nn.Sequential(*layers)


def transform(x):
    return np.sign(x) * np.log(np.abs(x) + 1)


def transform_inv(x):
    return np.sign(x) * (np.exp(np.abs(x)) - 1)


def load_data_pq(file_name_in):
    table = pq.read_table(file_name_in)
    df = table.to_pandas()
    return df.to_numpy(), list(df.columns)


def top_tr_types(np_data, tr_types_col, tr_amounts_col, f):
    tr_amounts_by_value = {}
    for i in range(len(np_data)):
        tr_types_ = np_data[i][tr_types_col]
        tr_amounts_ = f(np_data[i][tr_amounts_col])
        for ix_, j in enumerate(tr_types_):
            if j in tr_amounts_by_value:
                tr_amounts_by_value[j] += tr_amounts_[ix_]
            else:
                tr_amounts_by_value[j] = tr_amounts_[ix_]

    amounts_by_val_sorted = sorted(tr_amounts_by_value.items(),
                                   key = lambda x: abs(x[1]), reverse = True)

    negative_items = []
    positive_items = []
    for pair in amounts_by_val_sorted:
        if pair[1] >= 0:
            positive_items += [pair[0]]
        else:
            negative_items += [pair[0]]
    return negative_items, positive_items


def get_distributions(np_data, tr_amounts_col, tr_types_col=None,
                      negative_items=None, positive_items=None, top_thr=None,
                      take_first_fraction=0, f=lambda x: x):
    set_top_neg_types = set(negative_items[:top_thr])
    set_top_pos_types = set(positive_items[:top_thr])

    sums_of_negative_target = [0 for _ in range(len(np_data))]
    sums_of_positive_target = [0 for _ in range(len(np_data))]

    neg_distribution = [[] for _ in range(len(np_data))]
    pos_distribution = [[] for _ in range(len(np_data))]

    for i in range(len(np_data)):
        num = len(np_data[i][tr_amounts_col])
        thr_target_ix = int(num * take_first_fraction)
        amount_target = f(np_data[i][tr_amounts_col][thr_target_ix:])

        tr_type_target = np_data[i][tr_types_col][thr_target_ix:]
        neg_tr_amounts_target = {}
        others_neg_tr_amounts_target = 0
        pos_tr_amounts_target = {}
        others_pos_tr_amounts_target = 0

        # create value counts for each top transaction type and create neg and pos sum for id

        for ixx, el in enumerate(tr_type_target):
            if amount_target[ixx] < 0:
                sums_of_negative_target[i] += amount_target[ixx]
            else:
                sums_of_positive_target[i] += amount_target[ixx]

            if el in set_top_neg_types:
                neg_tr_amounts_target[el] = neg_tr_amounts_target.get(el, 0) + amount_target[ixx]
            elif el in set_top_pos_types:
                pos_tr_amounts_target[el] = pos_tr_amounts_target.get(el, 0) + amount_target[ixx]
            elif amount_target[ixx] < 0:
                others_neg_tr_amounts_target += amount_target[ixx]
            elif amount_target[ixx] >= 0:
                others_pos_tr_amounts_target += amount_target[ixx]
            else:
                assert False, "Should not be!"

        # collect neg and pos distribution for id
        for j in negative_items[:top_thr]:
            if j in neg_tr_amounts_target:
                p_neg = neg_tr_amounts_target[j] / sums_of_negative_target[i]
            else:
                p_neg = 0.
            neg_distribution[i] += [p_neg]
        if sums_of_negative_target[i] != 0:
            neg_distribution[i] += [others_neg_tr_amounts_target / sums_of_negative_target[i]]
        else:
            neg_distribution[i] += [0.]

        for j in positive_items[:top_thr]:
            if j in pos_tr_amounts_target:
                p_pos = pos_tr_amounts_target[j] / sums_of_positive_target[i]
            else:
                p_pos = 0.
            pos_distribution[i] += [p_pos]
        if sums_of_positive_target[i] != 0:
            pos_distribution[i] += [others_pos_tr_amounts_target / sums_of_positive_target[i]]
        else:
            pos_distribution[i] += [0.]

    return sums_of_negative_target, sums_of_positive_target, neg_distribution, pos_distribution




import torch


class MatrixMasker:
    """
    Returns matrix masked with zeros for
    summing only positive pairs distances
    (e.g. (2k-1, 2k) and (2k, 2k - 1) for k in 1 to N).
    """
    def get_masked_matrix(self, matrix, classes):
        mask = (classes.unsqueeze(1) == classes.unsqueeze(0)).int()
        mask.fill_diagonal_(0)
        masked_matrix = matrix*mask
        return masked_matrix



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class IdFilter(IterableProcessingDataset):
    def __init__(self, id_col, relevant_ids):
        """Remove records which are not in `relevant_ids`

        Args:
            id_col: field where id is stored
            relevant_ids: list of ids which should be kept
        """
        super().__init__()

        self._id_col = id_col
        self._relevant_ids = set(relevant_ids)
        one_element = next(iter(self._relevant_ids))
        self._id_type = type(one_element)

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            _id = features[self._id_col]
            if not self._is_in_relevant_ids_with_type(_id):
                continue
            yield rec

    def _is_in_relevant_ids_with_type(self, _id):
        if type(_id) is not self._id_type:
            raise TypeError(f'Type mismatch when id check. {type(_id)} found in sequence, '
                            f'but {self._id_type} from relevant_ids expected')

        return _id in self._relevant_ids



import torch

from ptls.loss import ZILNLoss


def test_ziln_loss():
    ziln, B = ZILNLoss(), 10
    assert ziln(torch.randn(B, 3), torch.randn(B, 1)) >= 0
    assert ziln(torch.randn(B, 3 + 3), torch.randn(B, 3)) >= 0


from .augmentation_dataset import AugmentationDataset
from .persist_dataset import PersistDataset
from .memory_dataset import MemoryMapDataset, MemoryIterableDataset
from .parquet_dataset import ParquetFiles, ParquetDataset
from .parquet_file_scan import parquet_file_scan
from .dataloaders import inference_data_loader



import torch
import numpy as np

from omegaconf import OmegaConf
from ptls.custom_layers import DropoutEncoder, Squeeze, CatLayer, MLP, TabularRowEncoder, CombinedTargetHeadFromRnn


class TrxEncoderTest(torch.nn.Module):
    def forward(self, x):
        return x


def test_dropout_encoder():
    drop_encoder = DropoutEncoder(p=0.5)
    x = torch.rand(256, 100)

    drop_encoder.eval()
    assert torch.equal(x, drop_encoder(x))

    drop_encoder.train()
    assert x.shape == drop_encoder(x).shape


def test_squeeze():
    x = torch.rand(256, 100, 1)
    squeeze = Squeeze()
    y = squeeze(x)
    assert y.shape == (256, 100)


def test_mlp():
    mlp_config = {
        "hidden_layers_size": [512, 100],
        "drop_p": 0.5,
        "objective": "classification"
    }
    mlp_config = OmegaConf.create(mlp_config)

    mlp = MLP(512, mlp_config)
    x = torch.rand(256, 512)
    y = mlp(x)
    assert y.shape == (256,)


def test_cat_layer():
    left_tail = torch.nn.Linear(100, 10)
    right_tail = torch.nn.Linear(200, 20)

    cat_layer = CatLayer(left_tail, right_tail)
    l = torch.rand(256, 100)
    r = torch.rand(256, 200)
    x = (l, r)
    y = cat_layer(x)
    assert y.shape == (256, 30)


def test_embedding_generator():
    tabular_config = {
        'num_features_count': 10,
        'cat_features_dims': [10, 10],
        'cat_emb_dim': 4
    }

    tabular_row_encoder = TabularRowEncoder(
        input_dim=tabular_config['num_features_count'] + len(tabular_config['cat_features_dims']),
        cat_dims=tabular_config['cat_features_dims'],
        cat_idxs=[x + tabular_config['num_features_count'] for x in range(len(tabular_config['cat_features_dims']))],
        cat_emb_dim=tabular_config['cat_emb_dim']
    )

    assert tabular_row_encoder.output_size == 18

    num = torch.rand(256, 10)
    cat = torch.randint(0, 10, (256, 2))
    x = torch.cat([num, cat], dim=1)
    y = tabular_row_encoder(x)

    assert y.shape == (256, 18)


def test_distribution_target_head():
    distr_target_head_config = {
        "in_size": 48,
        "num_distr_classes_pos": 4,
        "num_distr_classes_neg": 14,
        'use_gates': False,
        'pass_samples': False
    }

    distr_target_head = CombinedTargetHeadFromRnn(**distr_target_head_config)
    x = torch.rand(64, 48)
    x = (x, np.ones((64,)), np.ones((64,)))
    y = distr_target_head(x)
    assert type(y) == dict and len(y) == 4
    assert y['neg_sum'].shape == y['pos_sum'].shape == (64, 1) and \
                    y['neg_distribution'].shape == (64, 14) and y['pos_distribution'].shape == (64, 4)



import torch

from ptls.data_load.padded_batch import PaddedBatch


class RBatchNorm(torch.nn.Module):
    def __init__(self, num_features):
        super().__init__()

        self.bn = torch.nn.BatchNorm1d(num_features)

    def forward(self, v: PaddedBatch):
        x = v.payload
        B, T, H = x.size()  # B x T X H
        x = x.view(B * T, H)
        x = self.bn(x)
        x = x.view(B, T, H)
        return PaddedBatch(x, v.seq_lens)


class RBatchNormWithLens(torch.nn.Module):
    """
    The same as RBatchNorm, but ...
    Drop padded symbols (zeros) from batch when batch stat update
    """
    def __init__(self, num_features):
        super().__init__()

        self.bn = torch.nn.BatchNorm1d(num_features)

    def forward(self, v: PaddedBatch):
        x = v.payload
        B, T, H = x.size()  # B x T X H

        mask = v.seq_len_mask.bool()
        x_new = x.clone()
        x_new[mask] = self.bn(x[mask])
        return PaddedBatch(x_new, v.seq_lens)



import pickle

import torch

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.trx_encoder import TrxEncoderOhe


def test_example():
    B, T = 5, 20
    trx_encoder = TrxEncoderOhe(
        embeddings={'mcc_code': {'in': 100}},
        numeric_values={'amount': 'log'},
    )
    x = PaddedBatch(
        payload={
            'mcc_code': torch.randint(1, 99, (B, T)),
            'amount': torch.randn(B, T),
        },
        length=torch.randint(10, 20, (B,)),
    )
    z = trx_encoder(x)
    assert z.payload.shape == (5, 20, 100)  # B, T, H
    assert trx_encoder.output_size == 100
    z_sum = z.payload[:, :, :-1].sum(dim=2)
    assert (z_sum == 1).all()



import logging
from itertools import islice
import numpy as np
from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


logger = logging.getLogger(__name__)


class IterableShuffle(IterableProcessingDataset):
    def __init__(self, buffer_size):
        """

        Args:
            buffer_size: buffer size in records
        """
        super().__init__()

        assert buffer_size > 1  # we will split buffer into two parts
        self._buffer_size = buffer_size

    def __iter__(self):
        source = iter(self._src)
        buffer = np.array([])
        while True:
            new_buffer_size = self._buffer_size - len(buffer)
            new_buffer_list = list(islice(source, new_buffer_size))
            new_buffer = np.empty(len(new_buffer_list), dtype=object)
            new_buffer[:] = new_buffer_list
            # print(f'Buffer shape: {buffer.shape}, {new_buffer.shape}')
            buffer = np.concatenate([buffer, new_buffer])
            if len(buffer) == 0:
                break

            window_size = min(len(buffer), self._buffer_size // 2)
            ix_for_choice = np.random.choice(len(buffer), window_size, replace=False)

            for ix in ix_for_choice:
                yield buffer[ix]

            mask_selected = np.zeros(len(buffer), dtype=np.bool)
            mask_selected[ix_for_choice] = True
            buffer = buffer[~mask_selected]



import itertools
import logging
import warnings

import numpy as np
from sklearn.metrics import cohen_kappa_score

logger = logging.getLogger(__name__)


def block_iterator(iterator, size):
    bucket = list()
    for e in iterator:
        bucket.append(e)
        if len(bucket) >= size:
            yield bucket
            bucket = list()
    if bucket:
        yield bucket


def cycle_block_iterator(iterator, size):
    return block_iterator(itertools.cycle(iterator), size)


class ListSubset:
    """The same as torch.utils.data.Subset
    """
    def __init__(self, delegate, idx_to_take):
        self.delegate = delegate
        self.idx_to_take = idx_to_take

    def __len__(self):
        return len(self.idx_to_take)

    def __getitem__(self, idx):
        return self.delegate[self.idx_to_take[idx]]

    def __iter__(self):
        for i in self.idx_to_take:
            yield self.delegate[i]


def eval_kappa_regression(y_true, y_pred):
    dist = {3: 0.5, 0: 0.239, 2: 0.125, 1: 0.136} # ground shares

    acum = 0
    bound = {}
    for i in range(3):
        acum += dist[i]
        bound[i] = np.percentile(y_pred, acum * 100)

    y_pred = np.where(
        y_pred<=bound[0],
        0,
        np.where(
            y_pred<=bound[1],
            1,
            np.where(
                y_pred<=bound[2],
                2,
                3
            )
        )
    ).reshape(y_true.shape)

    return cohen_kappa_score(y_true, y_pred, weights='quadratic')


class Deprecated:
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""
    def __init__(self, message):
        self._message = message

    def __call__(self, func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn(f"Call to deprecated function {func.__name__}. {self._message}",
                          category=DeprecationWarning,
                          stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return func(*args, **kwargs)
        return new_func



import logging
import warnings

import torch

from torch.utils.data import DataLoader
from typing import List, Dict

from ptls.data_load import IterableChain
from ptls.data_load.augmentations.build_augmentations import build_augmentations
from ptls.data_load.data_module.coles_data_module import coles_collate_fn
from ptls.data_load.datasets import MemoryMapDataset
from ptls.data_load.iterable_processing.feature_filter import FeatureFilter
from ptls.data_load.iterable_processing.filter_non_array import FilterNonArray
from ptls.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from ptls.data_load.iterable_processing.to_torch_tensor import ToTorch
from ptls.frames.coles import split_strategy
from ptls.metric_learn.dataset.splitting_dataset import MapSplittingDataset

logger = logging.getLogger(__name__)


class PostProcessDataset(torch.utils.data.Dataset):
    def __init__(self, parent, post_processor):
        warnings.warn('Use `ptls.data_load.datasets.MemoryMapDataset`')
        super().__init__()
        self.parent = parent
        self.post_processor = post_processor

    def __len__(self):
        return len(self.parent)

    def __getitem__(self, idx):
        return next(self.post_processor([self.parent[idx]]))


r"""Generate a train data loader

Parameters
----------
data: List[Dict]
The dataset
min_seq_len: int. Default: 0.
The minimal length of sequences used for training. The shorter sequences would be skipped.
seq_split_strategy: str. Options: 'SampleSlices',  Default: 'SampleSlices'.
The strategy for splitting sequence to subsequences
split_count: int. Default: 5.
An amount of subsequences from one sequence.
split_cnt_min: int. Default: 1.
The minimal amount of events in each subsequence.
split_cnt_max: int. Default: 4000.
    The maximal amount of events in each subsequence.
num_workers: int. Default: 0.
The number of workers for the dataloader. 0 = single-process loader
train_batch_size: int. Default: 512.
The number of samples (before splitting to subsequences) in each batch
"""
def train_data_loader(
    data: List[Dict],
    min_seq_len: int = 0,
    seq_split_strategy: str = 'SampleSlices',
    split_count: int = 5,
    split_cnt_min: int = 1,
    split_cnt_max: int = 4000,
    drop_cols: List[str] = list(),
    num_workers: int = 0,
    batch_size: int = 512):
    warnings.warn('Use `train_dataloader()` method of `ptls.frames.PtlsDataModule` '
                  'with `ptls.frames.coles.ColesDataset` or `ptls.frames.coles.ColesIterableDataset`',
                  DeprecationWarning)

    dataset = MemoryMapDataset(
        data,
        i_filters=[
            SeqLenFilter(min_seq_len=min_seq_len),
            FilterNonArray(),
            ToTorch(),
            FeatureFilter(drop_feature_names=drop_cols),
        ],
    )

    split_strategy_params = {
        'split_strategy': seq_split_strategy,
        'split_count': split_count,
        'cnt_min': split_cnt_min,
        'cnt_max': split_cnt_max
    }

    dataset = MapSplittingDataset(
        base_dataset=dataset,
        splitter=split_strategy.create(**split_strategy_params),
        a_chain=build_augmentations({}),
    )

    return DataLoader(
        dataset=dataset,
        collate_fn=coles_collate_fn,
        shuffle=False,
        num_workers=num_workers,
        batch_size=batch_size
    )



import pytorch_lightning as pl
import torch.optim

from ptls.frames.coles import CoLESModule
from ptls_tests.test_data_load import RandomEventData
from pyhocon import ConfigFactory
from ptls.nn.seq_encoder import RnnSeqEncoder
from ptls.nn import Head
from ptls.nn import TrxEncoder
from functools import partial


def tst_params():
    params = {
        "data_module": {
            "train": {
                "num_workers": 1,
                "batch_size": 32,
                "trx_dropout": 0.01,
                "max_seq_len": 100,
            },
            "valid": {
                "batch_size": 16,
                "num_workers": 1,
                "max_seq_len": 100
            }
        },
        "rnn": {
            "type": "gru",
            "hidden_size": 16,
            "bidir": False,
            "trainable_starter": "static"
        },
        "trx_encoder": {
            "embeddings_noise": .003,
            "norm_embeddings": False,
            'embeddings': {
                'mcc_code': {'in': 21, 'out': 3},
                'trans_type': {'in': 11, 'out': 2},
            },
            'numeric_values': {'amount': 'log'}
        },
    }

    params = ConfigFactory.from_dict(params)
    return params


def test_train_loop():
    params = tst_params()

    model = CoLESModule(
        seq_encoder=RnnSeqEncoder(
            trx_encoder=TrxEncoder(**params['trx_encoder']),
            **params['rnn'],
        ),
        head=Head(use_norm_encoder=True),
        optimizer_partial=partial(torch.optim.Adam),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=1, gamma=1.0),
    )
    dl = RandomEventData(params['data_module'])
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)



from typing import Union
from typing import Dict

import torch
from torch import nn as nn

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.trx_encoder.scalers import scaler_by_name, BaseScaler


class TrxEncoderBase(nn.Module):
    """Base class for TrxEncoders.

    This provides a parts of embeddings for feature fields.
    This doesn't provide a full embedding of transaction.

    Parameters
    ----------
    embeddings:
        dict with categorical feature names.
        Values must be like this `{'in': dictionary_size, 'out': embedding_size}`
        These features will be encoded with lookup embedding table of shape (dictionary_size, embedding_size)
        Values can be a `torch.nn.Embedding` implementation
    numeric_values:
        dict with numerical feature names.
        Values must be a string with scaler_name.
        Possible values are: 'identity', 'sigmoid', 'log', 'year'.
        These features will be scaled with selected scaler.
        Values can be `ptls.nn.trx_encoder.scalers.BaseScaler` implementation

        One field can have many scalers. In this case key become alias and col name should be in scaler.
        Example:
        >>> TrxEncoderBase(
        >>>     numeric_values={
        >>>         'amount_orig': IdentityScaler(col_name='amount'),
        >>>         'amount_log': LogScaler(col_name='amount'),
        >>>     },
        >>> )

    out_of_index:
        How to process a categorical indexes which are greater than dictionary size.
        'clip' - values will be collapsed to maximum index. This works well for frequency encoded categories.
            We join infrequent categories to one.
        'assert' - raise an error of invalid index appear.
    """
    def __init__(self,
                 embeddings: Dict[str, Union[Dict, torch.nn.Embedding]] = None,
                 numeric_values: Dict[str, Union[str, BaseScaler]] = None,
                 out_of_index: str = 'clip',
                 ):
        super().__init__()

        if embeddings is None:
            embeddings = {}
        if numeric_values is None:
            numeric_values = {}

        self.embeddings = torch.nn.ModuleDict()
        for col_name, emb_props in embeddings.items():
            if type(emb_props) is dict:
                if emb_props.get('disabled', False):
                    continue
                if emb_props['in'] == 0 or emb_props['out'] == 0:
                    continue
                if emb_props['in'] < 3:
                    raise AttributeError(f'At least 3 should be in `embeddings.{col_name}.in`. '
                                         f'0-padding and at least two different embedding indexes')
                self.embeddings[col_name] = torch.nn.Embedding(
                    num_embeddings=emb_props['in'],
                    embedding_dim=emb_props['out'],
                    padding_idx=0,
                )
            elif isinstance(emb_props, torch.nn.Embedding):
                self.embeddings[col_name] = emb_props
            else:
                raise AttributeError(f'Wrong type of embeddings, found {type(col_name)} for "{col_name}"')
        self.out_of_index = out_of_index
        assert out_of_index in ('clip', 'assert')

        self.numeric_values = torch.nn.ModuleDict()
        for col_name, scaler_name in numeric_values.items():
            if type(scaler_name) is str:
                if scaler_name == 'none':
                    continue
                self.numeric_values[col_name] = scaler_by_name(scaler_name)
            elif isinstance(scaler_name, BaseScaler):
                self.numeric_values[col_name] = scaler_name
            else:
                raise AttributeError(f'Wrong type of numeric_values, found {type(scaler_name)} for "{col_name}"')

    def get_category_indexes(self, x: PaddedBatch, col_name: str):
        """Returns category feature values clipped to dictionary size.

        Parameters
        ----------
        x: PaddedBatch with feature dict. Each value is `(B, T)` size
        col_name: required feature name
        """
        v = x.payload[col_name].long()
        max_size = self.embeddings[col_name].num_embeddings

        if self.out_of_index == 'clip':
            return v.clip(0, max_size - 1)
        if self.out_of_index == 'assert':
            out_of_index_cnt = (v >= max_size).sum()
            if out_of_index_cnt > 0:
                raise IndexError(f'Found indexes greater than dictionary size for "{col_name}"')
            return v
        raise AssertionError(f'Unknown out_of_index value: {self.out_of_index}')

    def get_category_embeddings(self, x: PaddedBatch, col_name: str):
        indexes = self.get_category_indexes(x, col_name)
        return self.embeddings[col_name](indexes)

    def get_numeric_scaled(self, x: PaddedBatch, col_name):
        """Returns numerical feature values transformed with selected scaler.

        Parameters
        ----------
        x: PaddedBatch with feature dict. Each value is `(B, T)` size
        col_name: required feature name
        """
        scaler = self.numeric_values[col_name]
        if scaler.col_name is None:
            v = x.payload[col_name].unsqueeze(2).float()
        else:
            v = x.payload[scaler.col_name].unsqueeze(2).float()
        return scaler(v)

    @property
    def numerical_size(self):
        return sum(n.output_size for n in self.numeric_values.values())

    @property
    def embedding_size(self):
        return sum(e.embedding_dim for e in self.embeddings.values())

    @property
    def output_size(self):
        s = self.numerical_size + self.embedding_size
        return s

    @property
    def category_names(self):
        """Returns set of used feature names
        """
        return set([field_name for field_name in self.embeddings.keys()] +
                   [value_name for value_name in self.numeric_values.keys()]
                   )

    @property
    def category_max_size(self):
        """Returns dict with categorical feature names. Value is dictionary size
        """
        return {k: v['in'] for k, v in self.embeddings.items()}



import torch

from ptls.frames.coles.losses.complex_loss import ComplexLoss
from ptls.frames.coles.losses import (
    MarginLoss, HistogramLoss, TripletLoss, BinomialDevianceLoss, ContrastiveLoss,
    CentroidLoss, CentroidSoftmaxLoss, SoftmaxLoss, SoftmaxPairwiseLoss
)
from ptls.frames.coles.sampling_strategies import AllPositivePairSelector, AllTripletSelector, MatrixMasker, PairwiseMatrixSelector


def get_data():
    B, C, H = 2, 2, 2  # Batch, num Classes, Hidden size
    x = torch.tensor([
        [0.9611,  0.2761],
        [0.6419,  0.7668],
        [-0.7595,  0.6505],
        [-0.3806,  0.9247],
    ])
    # x = torch.randn(B * C, H)
    x = x.div(x.pow(2).sum(dim=1, keepdim=True).pow(0.5))
    y = torch.arange(C).view(-1, 1).expand(C, B).reshape(-1)
    return x, y


def test_contrastive_loss():
    x, y = get_data()
    sampling_strategy = AllPositivePairSelector()
    loss_fn = ContrastiveLoss(0.5, sampling_strategy)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_binomial_deviance_loss():
    x, y = get_data()
    sampling_strategy = AllPositivePairSelector()
    loss_fn = BinomialDevianceLoss(sampling_strategy)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_margin_loss():
    x, y = get_data()
    sampling_strategy = AllPositivePairSelector()
    loss_fn = MarginLoss(sampling_strategy)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_triplet_loss():
    x, y = get_data()
    sampling_strategy = AllTripletSelector()
    loss_fn = TripletLoss(0.5, sampling_strategy)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_histogram_loss1():
    x, y = get_data()
    loss_fn = HistogramLoss(num_steps=5)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_histogram_loss2():
    x, y = get_data()
    loss_fn = HistogramLoss(num_steps=51)

    print(x)
    print(y)
    loss = loss_fn(x, y)
    print(loss)
    assert 1 == 1


def test_complex_loss():
    B, C, H = 2, 2, 2  # Batch, num Classes, Hidden size
    x_ml = torch.randn(B * C, H)
    x_ml = x_ml.div(x_ml.pow(2).sum(dim=1, keepdim=True).pow(0.5))
    y_ml = torch.arange(C).view(-1, 1).expand(C, B).reshape(-1, 1)

    sampling_strategy = AllPositivePairSelector()
    ml_loss = MarginLoss(sampling_strategy)

    x_aug = torch.randn(B * C, C)
    x_aug = x_aug.div(x_aug.sum(dim=1, keepdim=True))
    y_aug = torch.arange(B).view(1, -1).expand(C, B).reshape(-1, 1)
    aug_loss_fn = torch.nn.NLLLoss()

    y = torch.cat([y_ml, y_aug], axis=1)
    loss_fn = ComplexLoss(ml_loss, aug_loss_fn, 0.5)
    loss = loss_fn((x_aug, x_ml), y)
    print(loss)
    assert 1 == 1


def test_centroid_loss():
    x, y = get_data()
    loss_fn = CentroidLoss(class_num=None)

    loss = loss_fn(x, y)
    assert loss > 0


def test_centroid_loss_class_num():
    x = torch.tensor([
        [0, 0],  # 1, 1
        [2, 2],  # 1, 1
        [-1, 0],  # 0, -1
        [1, -2],  # 0, -1
    ])
    y = torch.tensor([0, 0, 1, 1])
    loss_fn = CentroidLoss(class_num=2)

    loss = loss_fn(x, y)
    assert loss == 2


def test_centroid_softmax_loss():
    x, y = get_data()
    loss_fn = CentroidSoftmaxLoss(class_num=None, temperature=1.0)

    loss = loss_fn(x, y)
    assert loss > 0


def test_softmax_loss():
    x = torch.tensor([
        [1.0, 3.0],
        [2.5, 2.2],
        [-1.0, 0.75],
        [1.25, -2.0],
    ])
    y = torch.tensor([0, 0, 1, 1])

    loss_fn = SoftmaxLoss(masker=MatrixMasker(), temperature=1.0)
    loss = loss_fn(x, y)

    true_value = 1.79
    assert abs(true_value - loss) < 1e-3
    assert type(loss) is torch.Tensor


def test_softmax_pairwise_loss():
    x = torch.tensor([
        [1.0, 3.0],
        [2.0, 2.0],
        [-1.0, 1.0],
        [3.0, -2.0],
        [4.0, 0.0],
        [-2.0, 5.0],
    ])
    y = torch.tensor([0, 0, 0, 1, 1, 1])

    loss_fn = SoftmaxPairwiseLoss(pair_selector=PairwiseMatrixSelector(), temperature=1.0)
    loss = loss_fn(x, y)

    true_value = 1.518
    assert abs(true_value - loss) < 1e-3
    assert type(loss) is torch.Tensor



import ast, glob, json, random, warnings
import logging

import pytorch_lightning as pl
import numpy as np
import pandas as pd
from embeddings_validation.file_reader import TargetFile
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from sklearn.model_selection import train_test_split

from ptls.data_load import padded_collate, padded_collate_distribution_target, IterableChain, IterableAugmentations
from ptls.data_load.augmentations.build_augmentations import build_augmentations
from ptls.data_load.data_module.map_augmentation_dataset import MapAugmentationDataset
from ptls.data_load.iterable_processing import FeatureFilter, FeatureTypeCast, IdFilter, IterableShuffle, SeqLenFilter, TargetJoin, TargetExtractor
from ptls.data_load.datasets.parquet_dataset import ParquetFiles, ParquetDataset
from ptls.data_load.utils import collate_target

logger = logging.getLogger(__name__)


class ClsDataModuleTrain(pl.LightningDataModule):
    def __init__(self,
                 type,
                 setup,
                 train,
                 valid,
                 pl_module,
                 fold_id,
                 test=None,
                 distribution_target_task=False,
                 distribution_target_size=None,
                 ):
        warnings.warn('Use `ptls.frames.PtlsDataModule` with `ptls.frames.supervised.SeqToTargetDataset` '
                      'or `ptls.frames.supervised.SeqToTargetIterableDataset`',
                      DeprecationWarning)
        super().__init__()

        self.fold_id = fold_id
        self._type = type
        assert self._type in ('map', 'iterable')

        self.distribution_target_task = distribution_target_task
        if distribution_target_size is None:
            self.y_function = int if not self.distribution_target_task else \
                lambda x: np.array(ast.literal_eval(x), dtype=object)
        elif distribution_target_size == 0:
            self.y_function = np.float32
        elif isinstance(distribution_target_size, int):
            self.y_function = lambda x: collate_target(ast.literal_eval(x), distribution_target_size)
        else:
            raise AttributeError(f"Undefined distribution target size: {distribution_target_size}")

        self.setup_conf = setup
        self.train_conf = train
        self.valid_conf = valid
        self.test_conf = test if test else self.valid_conf

        self.col_id = self.setup_conf.col_id
        self.col_id_dtype = {
            'str': str,
            'int': int,
        }[self.setup_conf.col_id_dtype]
        self.col_target = self.setup_conf.col_target
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self.test_dataset = None
        self.predict_dataset = None

        self._train_targets = None
        self._valid_targets = None
        self._test_targets = None
        self._predict_targets = None

        self._fold_info = None
        self.data_is_prepared = False
        self.do_test = "test_data_path" in setup.dataset_files

    def prepare_data(self):
        if not self.data_is_prepared:
            if 'dataset_files' in self.setup_conf:
                self.setup_iterable_files()
            elif 'dataset_parts' in self.setup_conf:
                self.setup_iterable_parts()
            else:
                raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')
            if self._type == 'map':
                self.setup_map()
            self.data_is_prepared = True

    def setup_iterable_files(self):
        if self.setup_conf.get('split_by', None) == 'embeddings_validation':

            if (self.setup_conf.get('use_files_partially', None)):
                n_train = len(glob.glob(self.setup_conf.dataset_files.train_data_path + "/*.parquet"))
                ixes = list(range(n_train))
                train_ixes, test_ixes = train_test_split(ixes, test_size=int(n_train * (1 - self.setup_conf.train_part)), shuffle=True)
                train_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path, train_ixes).data_files
                if self.setup_conf.same_file_for_test:
                    test_ixes = random.sample(test_ixes, int(n_train * self.setup_conf.test_part))
                    test_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path, test_ixes).data_files
                else:
                    n_test = len(glob.glob(self.setup_conf.dataset_files.test_data_path + "/*.parquet"))
                    test_ixes = random.sample(ixes, int(n_test * self.setup_conf.test_part))
                    test_data_files = ParquetFiles(self.setup_conf.dataset_files.test_data_path, test_ixes).data_files
            else:
                train_data_files = ParquetFiles(self.setup_conf.dataset_files.train_data_path).data_files
                test_data_files = ParquetFiles(self.setup_conf.dataset_files.test_data_path).data_files if self.do_test else []

            self.read_external_splits()
            self.train_dataset = ParquetDataset(
                train_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = ParquetDataset(
                train_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )
            if self.do_test:
                self.test_dataset = ParquetDataset(
                    test_data_files,
                    post_processing=IterableChain(*self.build_iterable_processing('test')),
                    shuffle_files=False,
                )
            self.predict_dataset = ParquetDataset(
                train_data_files + test_data_files,
                post_processing=IterableChain(*self.build_iterable_processing('predict')),
                shuffle_files=False,
            )

        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def setup_iterable_parts(self):
        raise NotImplementedError()

    def read_external_splits(self):
        if self.setup_conf.split_by == 'embeddings_validation':
            with open(self.setup_conf.fold_info, 'r') as f:
                self._fold_info = json.load(f)

        current_fold = self._fold_info[self.fold_id]
        self._train_targets = TargetFile.load(current_fold['train']['path']).df
        self._valid_targets = TargetFile.load(current_fold['valid']['path']).df
        if self.do_test:
            self._test_targets = TargetFile.load(current_fold['test']['path']).df
            self._predict_targets = pd.concat((self._valid_targets, self._test_targets))
        else:
            self._predict_targets = self._valid_targets

        labeled_amount = self.train_conf.get('labeled_amount', None)
        if labeled_amount is not None:
            _len_orig = len(self._train_targets)
            if type(labeled_amount) is float:
                labeled_amount = int(_len_orig * labeled_amount)
            self._train_targets = self._train_targets.iloc[:labeled_amount]
            logger.info(f'Reduced train amount from {_len_orig} to {len(self._train_targets)}')

    def build_iterable_processing(self, part):
        yield FeatureTypeCast({self.col_id: self.col_id_dtype})

        if 'dataset_files' in self.setup_conf and self.setup_conf.split_by == 'embeddings_validation':
            if part == 'train':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._train_targets[self.col_id].values.tolist())
            elif part == 'valid':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._valid_targets[self.col_id].values.tolist())
            elif part == 'test':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._test_targets[self.col_id].values.tolist())
            elif part == 'predict':
                yield IdFilter(id_col=self.col_id, relevant_ids=self._predict_targets[self.col_id].values.tolist())
            else:
                raise AttributeError(f'Unknown part: {part}')

        if part == 'train':
            yield SeqLenFilter(min_seq_len=self.train_conf.min_seq_len)

        if part == 'train':
            yield TargetJoin(self.col_id, self._train_targets.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        elif part == 'valid':
            yield TargetJoin(self.col_id, self._valid_targets.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        elif part == 'test':
            yield TargetJoin(self.col_id, self._test_targets.set_index(self.col_id)[self.col_target].to_dict(), self.y_function)
        elif part == 'predict':
            yield TargetExtractor(self.col_id, drop_from_features=False)
        else:
            raise AttributeError(f'Unknown part: {part}')

        yield FeatureFilter(keep_feature_names=self.category_names,
                            drop_feature_names=self.setup_conf.get('drop_feature_names', None))

        if self._type == 'iterable':
            # all processing in single chain
            if part == 'train':
                yield IterableShuffle(self.train_conf.buffer_size)
            yield IterableAugmentations(self.build_augmentations(part))

    def build_augmentations(self, part):
        if part == 'train':
            return build_augmentations(self.train_conf.augmentations)
        elif part == 'valid':
            return build_augmentations(self.valid_conf.augmentations)
        elif part in ('test', 'predict'):
            return build_augmentations(self.test_conf.augmentations)

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_target_task else padded_collate,
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf.num_workers,
            batch_size=self.train_conf.batch_size,
            drop_last=self.train_conf.get('drop_last', False)
        )

    def setup_map(self):
        if self.train_dataset is not None:
            self.train_dataset = list(tqdm(iter(self.train_dataset)))
            logger.info(f'Loaded {len(self.train_dataset)} for train')
            self.train_dataset = MapAugmentationDataset(
                base_dataset=self.train_dataset,
                a_chain=self.build_augmentations('train'),
            )
        if self.valid_dataset is not None:
            self.valid_dataset = list(tqdm(iter(self.valid_dataset)))
            logger.info(f'Loaded {len(self.valid_dataset)} for valid')
            self.valid_dataset = MapAugmentationDataset(
                base_dataset=self.valid_dataset,
                a_chain=self.build_augmentations('valid'),
            )
        if self.test_dataset is not None:
            self.test_dataset = list(tqdm(iter(self.test_dataset)))
            logger.info(f'Loaded {len(self.test_dataset)} for test')
            self.test_dataset = MapAugmentationDataset(
                base_dataset=self.test_dataset,
                a_chain=self.build_augmentations('test'),
            )
        if self.predict_dataset is not None:
            self.predict_dataset = list(tqdm(iter(self.predict_dataset)))
            logger.info(f'Loaded {len(self.predict_dataset)} for predict')
            self.predict_dataset = MapAugmentationDataset(
                base_dataset=self.predict_dataset,
                a_chain=self.build_augmentations('predict'),
            )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_target_task else padded_collate,
            num_workers=self.valid_conf.num_workers,
            batch_size=self.valid_conf.batch_size,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.test_dataset,
            collate_fn=padded_collate_distribution_target if self.distribution_target_task else padded_collate,
            num_workers=self.test_conf.num_workers,
            batch_size=self.test_conf.batch_size,
        )

    def predict_dataloader(self):
        return DataLoader(
            dataset=self.predict_dataset,
            collate_fn=padded_collate,
            num_workers=self.test_conf.num_workers,
            batch_size=self.test_conf.batch_size,
        )



import torch


# TODO: update this, split into 3 functions for corresponding targets, remove 'feature_arrays'
def gen_trx_data(lengths, target_share=.5, target_type='bin_cls', use_feature_arrays_key=True):
    n = len(lengths)
    # TODO: replace `lengths` with `min_len`, `max_len`, `num_sample`
    if target_type == 'bin_cls':
        targets = (torch.rand(n) >= target_share).long()
    elif target_type == 'multi_cls':
        targets = torch.randint(0, 4, (n,))
    elif target_type == 'regression':
        targets = torch.randn(n)
    else:
        raise AttributeError(f'Unknown target_type: {target_type}')

    samples = list()
    for target, length in zip(targets, lengths):
        s = dict()
        s['trans_type'] = (torch.rand(length)*10 + 1).long()
        s['mcc_code'] = (torch.rand(length) * 20 + 1).long()
        s['amount'] = (torch.rand(length) * 1000 + 1).long()

        if use_feature_arrays_key:
            samples.append({'feature_arrays': s, 'target': target})
        else:
            s['target'] = target
            samples.append(s)

    return samples



import torch
import pytorch_lightning as pl
from functools import partial


class PtlsDataModule(pl.LightningDataModule):
    def __init__(self,
                 train_data=None,
                 train_batch_size=1,
                 train_num_workers=0,
                 train_drop_last=False,
                 valid_data=None,
                 valid_batch_size=None,
                 valid_num_workers=None,
                 valid_drop_last=False,
                 test_data=None,
                 test_batch_size=None,
                 test_num_workers=None,
                 test_drop_last=False,
                 ):

        super().__init__()
        self.save_hyperparameters(ignore=['train_data', 'valid_data', 'test_data'])
        
        if self.hparams.valid_num_workers is None:
            self.hparams.valid_num_workers = self.hparams.train_num_workers
        if self.hparams.test_num_workers is None:
            self.hparams.test_num_workers = self.hparams.valid_num_workers

        if self.hparams.valid_batch_size is None:
            self.hparams.valid_batch_size = self.hparams.train_batch_size
        if self.hparams.test_batch_size is None:
            self.hparams.test_batch_size = self.hparams.valid_batch_size

        if train_data is not None:
            self.train_dataloader = partial(self.train_dl, train_data)
        if valid_data is not None:
            self.val_dataloader = partial(self.val_dl, valid_data)
        if test_data is not None:
            self.test_dataloader = partial(self.test_dl, test_data)
            
    def train_dl(self, train_data):
        return torch.utils.data.DataLoader(
            dataset=train_data,
            collate_fn=train_data.collate_fn,
            shuffle=not isinstance(train_data, torch.utils.data.IterableDataset),
            num_workers=self.hparams.train_num_workers,
            batch_size=self.hparams.train_batch_size,
            drop_last=self.hparams.train_drop_last,
        )

    def val_dl(self, valid_data):
        return torch.utils.data.DataLoader(
            dataset=valid_data,
            collate_fn=valid_data.collate_fn,
            shuffle=False,
            num_workers=self.hparams.valid_num_workers,
            batch_size=self.hparams.valid_batch_size,
            drop_last=self.hparams.valid_drop_last,
        )

    def test_dl(self, test_data):
        return torch.utils.data.DataLoader(
            dataset=test_data,
            collate_fn=test_data.collate_fn,
            shuffle=False,
            num_workers=self.hparams.test_num_workers,
            batch_size=self.hparams.test_batch_size,
            drop_last=self.hparams.test_drop_last,
        )


import argparse
import datetime
import logging
import os
import pickle
from random import Random

import numpy as np
import pandas as pd

import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql import SparkSession
from pyspark.sql import Window


logger = logging.getLogger(__name__)


class DatasetConverter:
    def __init__(self):
        self.config = None

    def parse_args(self, args=None):
        parser = argparse.ArgumentParser()

        parser.add_argument('--data_path', type=os.path.abspath)
        parser.add_argument('--trx_files', nargs='+')
        parser.add_argument('--target_files', nargs='*', default=[])
        parser.add_argument('--target_as_array', action='store_true')

        parser.add_argument('--print_dataset_info', action='store_true')
        parser.add_argument('--sample_fraction', type=float, default=None)
        parser.add_argument('--col_client_id', type=str)
        parser.add_argument('--cols_event_time', nargs='+')

        parser.add_argument('--dict', nargs='*', default=[])
        parser.add_argument('--cols_category', nargs='*', default=[])
        parser.add_argument('--cols_log_norm', nargs='*', default=[])
        parser.add_argument('--col_target', nargs='*', default=[])
        parser.add_argument('--test_size', default='0.1')
        parser.add_argument('--salt', type=int, default=42)
        parser.add_argument('--max_trx_count', type=int, default=5000)

        parser.add_argument('--output_train_path', type=os.path.abspath)
        parser.add_argument('--output_test_path', type=os.path.abspath)
        parser.add_argument('--output_test_ids_path', type=os.path.abspath)
        parser.add_argument('--save_partitioned_data', action='store_true')
        parser.add_argument('--log_file', type=os.path.abspath)

        args = parser.parse_args(args)
        logger.info('Parsed args:\n' + '\n'.join([f'  {k:15}: {v}' for k, v in vars(args).items()]))
        self.config = args

    def spark_read_file(self, path):
        spark = SparkSession.builder.getOrCreate()

        ext = os.path.splitext(path)[1]
        if ext == '.csv':
            return spark.read.option("escape", "\"").csv(path, header=True)
        elif ext == '.parquet':
            return spark.read.parquet(path)
        else:
            raise AttributeError(f'Unknown extension "{ext}" for "{path}"')

    def path_to_file(self, file_name):
        return os.path.join(self.config.data_path, file_name)

    def load_source_data(self, trx_files):
        """
        :param trx_files:
        :return: spark.DataFrame with `event_time` column of float type
        """
        data = None
        for file in trx_files:
            file_path = self.path_to_file(file)
            df = self.spark_read_file(file_path)
            data = df if data is None else data.union(df)
            logger.info(f'Loaded {df.count()} rows from "{file_path}"')

        cnt = data.count()
        logger.info(f'Loaded {cnt} rows in total')

        cnt_in_partition = 100000
        data = data.repartition((cnt + cnt_in_partition - 1) // cnt_in_partition)
        return data

    def pd_hist(self, df, name, bins=10):
        # logger.info('pd_hist begin')
        # logger.info(f'sf = {self.config.sample_fraction}')
        data = df.select(name)
        if self.config.sample_fraction is not None:
            data = data.sample(fraction=self.config.sample_fraction)
        data = data.toPandas()[name]

        if data.dtype.kind == 'f':
            round_len = 1 if data.max() > bins + 1 else 2
            bins = np.linspace(data.min(), data.max(), bins + 1).round(round_len)
        elif np.percentile(data, 99) - data.min() > bins - 1:
            bins = np.linspace(data.min(), np.percentile(data, 99), bins).astype(int).tolist() + [int(data.max() + 1)]
        else:
            bins = np.arange(data.min(), data.max() + 2, 1).astype(int)
        df = pd.cut(data, bins, right=False).rename(name)
        df = df.to_frame().assign(cnt=1).groupby(name)[['cnt']].sum()
        df['% of total'] = df['cnt'] / df['cnt'].sum()
        return df

    def get_encoder(self, df, col_name):
        df = df.withColumn(col_name, F.coalesce(F.col(col_name).cast('string'), F.lit('#EMPTY')))

        col_orig = '_orig_' + col_name
        df = df.withColumnRenamed(col_name, col_orig)

        df_encoder = df.groupby(col_orig).agg(F.count(F.lit(1)).alias('_cnt'))
        df_encoder = df_encoder.withColumn(col_name,
                                           F.row_number().over(Window.partitionBy().orderBy(F.col('_cnt').desc())))
        df_encoder = df_encoder.withColumn(col_name, F.col(col_name) + F.lit(1))
        df_encoder = df_encoder.withColumn(col_name, F.col(col_name))
        df_encoder = df_encoder.drop('_cnt')

        df_encoder = df_encoder.repartition(1)
        df_encoder.persist()
        _ = df_encoder.count()

        return df_encoder

    def encode_col(self, df, col_name, df_encoder):
        df = df.withColumn(col_name, F.coalesce(F.col(col_name).cast('string'), F.lit('#EMPTY')))

        col_orig = '_orig_' + col_name
        df = df.withColumnRenamed(col_name, col_orig)

        df = df.join(df_encoder, on=col_orig, how='left')
        df = df.withColumn(col_name, F.coalesce(F.col(col_name), F.lit(1)))
        df = df.drop(col_orig)

        return df

    def log_transform(self, df, col_name):
        df = df.withColumn(col_name, F.coalesce(F.col(col_name), F.lit(0)))
        df = df.withColumn(col_name, F.signum(F.col(col_name)) * F.log(F.abs(F.col(col_name)) + F.lit(1)))
        return df

    def _td_default(self, df, cols_event_time):
        raise NotImplementedError()

    def _td_float(self, df, col_event_time):
        df = df.withColumn('event_time', F.col(col_event_time).astype('float'))
        logger.info('To-float time transformation')
        return df

    def _td_datetime(self, df, col_event_time):
        df = df.withColumn('event_time', F.unix_timestamp(F.col(col_event_time)) / F.lit(24 * 60 * 60))
        logger.info('Datetime-to-unix-timestamp time transformation')
        return df

    def _td_gender(self, df, col_event_time):
        """Gender-dataset-like transformation

        'd hh:mm:ss' -> float where integer part is day number and fractional part is seconds from day begin
        '1 00:00:00' -> 1.0
        '1 12:00:00' -> 1.5
        '1 01:00:00' -> 1 + 1 / 24
        '2 23:59:59' -> 1.99
        '432 12:00:00' -> 432.5   '000432 12:00:00'

        :param df:
        :param col_event_time:
        :return:
        """
        df = df.withColumn('_et_day', F.substring(F.lpad(F.col(col_event_time), 15, '0'), 1, 6).cast('float'))

        df = df.withColumn('_et_time', F.substring(F.lpad(F.col(col_event_time), 15, '0'), 8, 8))
        df = df.withColumn('_et_time', F.regexp_replace('_et_time', r'\:60$', ':59'))
        df = df.withColumn('_et_time', F.unix_timestamp('_et_time', 'HH:mm:ss') / (24 * 60 * 60))

        df = df.withColumn('event_time', F.col('_et_day') + F.col('_et_time'))
        df = df.drop('_et_day', '_et_time')
        logger.info('Gender-dataset-like time transformation')
        return df

    def remove_long_trx(self, df, max_trx_count, col_client_id):
        """
        This function select the last max_trx_count transactions
        """
        df = df.withColumn('_cn', F.count(F.lit(1)).over(Window.partitionBy(col_client_id)))
        df = df.withColumn('_rn', F.row_number().over(
            Window.partitionBy(col_client_id).orderBy(F.col('event_time').desc())))
        df = df.filter(F.col('_rn') <= max_trx_count)
        df = df.drop('_cn')
        df = df.drop('_rn')
        return df

    def collect_lists(self, df, col_id):
        col_list = [col for col in df.columns if col != col_id]

        if self.config.save_partitioned_data:
            df = df.withColumn('mon_id', (F.col('event_time') / 30).cast('int'))
            col_id = [col_id, 'mon_id']

        df = df \
            .withColumn('trx_count', F.count(F.lit(1)).over(Window.partitionBy(col_id))) \
            .withColumn('_rn', F.row_number().over(Window.partitionBy(col_id).orderBy('event_time')))

        for col in col_list:
            df = df.withColumn(col, F.collect_list(col).over(Window.partitionBy(col_id).orderBy('_rn'))) \

        df = df.filter('_rn = trx_count').drop('_rn')
        return df

    def join_dict(self, df, df_dict_name, col_id):
        path = self.path_to_file(df_dict_name)
        df_dict = self.spark_read_file(path)
        df = df.join(df_dict, on=col_id, how='left')

        col_counter = 0
        for col in df_dict.columns:
            if col == col_id:
                continue
            col_counter += 1
        logger.info(f'Join with "{path}" done. New {col_counter} columns joined')
        return df

    def trx_to_features(self, df_data, print_dataset_info,
                        col_client_id, cols_event_time, cols_category, cols_log_norm, max_trx_count):
        if print_dataset_info:
            unique_clients = df_data.select(col_client_id).distinct().count()
            logger.info(f'Found {unique_clients} unique clients')

        for col in cols_log_norm:
            df_data = self.log_transform(df_data, col)
            if print_dataset_info:
                logger.info(f'Encoder stat for "{col}":\ncodes | trx_count\n{self.pd_hist(df_data, col)}')

        encoders = {col: self.get_encoder(df_data, col) for col in cols_category}
        for col in cols_category:
            df_data = self.encode_col(df_data, col, encoders[col])
            if print_dataset_info:
                logger.info(f'Encoder stat for "{col}":\ncodes | trx_count\n{self.pd_hist(df_data, col)}')

        if print_dataset_info:
            df = df_data.groupby(col_client_id).agg(F.count(F.lit(1)).alias("trx_count"))
            logger.info(f'Trx count per clients:\nlen(trx_list) | client_count\n{self.pd_hist(df, "trx_count")}')

        # column filter
        used_columns = [col for col in df_data.columns
                        if col in cols_category + cols_log_norm + ['event_time', col_client_id]]

        logger.info('Feature collection in progress ...')
        features = df_data.select(used_columns)
        features = self.remove_long_trx(features, max_trx_count, col_client_id)
        features = self.collect_lists(features, col_client_id)

        if print_dataset_info:
            feature_names = list(features.columns)
            logger.info(f'Feature names: {feature_names}')

        features.persist()
        logger.info(f'Prepared features for {features.count()} clients')
        return features

    def update_with_target(self, features, df_target, col_client_id, col_target):
        if type(col_target) is list and self.config.target_as_array:
            col_list = []
            for col in col_target:
                col_list.append(F.col(col))
            df_target = df_target.withColumn("target", F.array(col_list)) 
            df_target = df_target.select(col_client_id, "target")
        elif type(col_target) is list and not self.config.target_as_array:
            col_list = []
            for col in col_target:
                if col.startswith('target'):
                    col_list.append(F.col(col))
                else:
                    col_list.append(F.col(col).alias(f'target_{col}'))
            df_target = df_target.select(col_client_id, *col_list)
        else:
            col_list = [F.col(col_client_id).alias(col_client_id)]
            col_list.append(F.col(col_target).cast('int').alias('target'))
            df_target = df_target.select(*col_list)
            df_target = df_target.repartition(1)

        features = features.join(df_target, on=col_client_id, how='left')
        features.persist()

        logger.info(f'Target updated for {features.count()} clients')
        return features

    def split_dataset(self, all_data, test_size, df_target, col_client_id, salt):
        spark = SparkSession.builder.getOrCreate()

        s_clients = set(cl[0] for cl in df_target.select(col_client_id).distinct().collect())

        # shuffle client list
        s_all_data_clients = set(cl[0] for cl in all_data.select(col_client_id).distinct().collect())
        s_clients = sorted(cl_id for cl_id in s_clients if cl_id in s_all_data_clients)
        Random(salt).shuffle(s_clients)

        # split client list
        Nrows_test = int(len(s_clients) * test_size)
        s_clients_train = s_clients[:-Nrows_test]
        s_clients_test = s_clients[-Nrows_test:]

        s_clients_train = spark.createDataFrame([(i,) for i in s_clients_train], [col_client_id]).repartition(1)
        s_clients_test = spark.createDataFrame([(i,) for i in s_clients_test], [col_client_id]).repartition(1)
        s_clients = spark.createDataFrame([(i,) for i in s_clients], [col_client_id]).repartition(1)

        # split data
        labeled_train = all_data.join(s_clients_train, on=col_client_id, how='inner')
        labeled_test = all_data.join(s_clients_test, on=col_client_id, how='inner')
        unlabeled = all_data.join(s_clients, on=col_client_id, how='left_anti')

        train = labeled_train.union(unlabeled)
        test = labeled_test

        logger.info(f'Train size: {train.count()} clients')
        logger.info(f'Test size: {test.count()} clients')

        return train, test

    def split_dataset_predefined(
                self,
                all_data,
                data_path,
                col_client_id,
                test_ids_path,
            ):
        df_test = self.load_source_data([test_ids_path])
        df_test = df_test.withColumn('_is_test', F.lit(1))

        all_data = all_data.join(df_test, on=col_client_id, how='left')
        all_data = all_data.withColumn('_is_test', F.coalesce(F.col('_is_test'), F.lit(0)))

        train = all_data.filter("_is_test = 0")
        test = all_data.filter("_is_test = 1")
        return train, test

    def save_features(self, df_data, save_path):
        if not self.config.save_partitioned_data:
            df_data.write.parquet(save_path, mode='overwrite')
            logger.info(f'Saved to: "{save_path}"')
        else:
            df_data = df_data.withColumn('hash_id', F.crc32(F.col(self.config.col_client_id)) % 100)
            df_data.write.parquet(save_path, mode='overwrite', partitionBy=['mon_id', 'hash_id'])
            logger.info(f'Saved partitions to: "{save_path}"')

    def run(self):
        _start = datetime.datetime.now()
        self.parse_args()
        spark = SparkSession.builder.getOrCreate()

        self.logging_config()

        # description
        spark.sparkContext.setLocalProperty('callSite.short', 'load_source_data')
        source_data = self.load_transactions()

        # description
        spark.sparkContext.setLocalProperty('callSite.short', 'trx_to_features')
        client_features = self.trx_to_features(
            df_data=source_data,
            print_dataset_info=self.config.print_dataset_info,
            col_client_id=self.config.col_client_id,
            cols_event_time=self.config.cols_event_time,
            cols_category=self.config.cols_category,
            cols_log_norm=self.config.cols_log_norm,
            max_trx_count=self.config.max_trx_count,
        )

        if len(self.config.target_files) > 0 and len(self.config.col_target) > 0:
            # load target
            df_target = self.load_target()
            df_target.persist()

            if len(self.config.col_target) == 1:
                col_target = self.config.col_target[0]
            else:
                col_target = self.config.col_target

            # description
            spark.sparkContext.setLocalProperty('callSite.short', 'update_with_target')
            client_features = self.update_with_target(
                features=client_features,
                df_target=df_target,
                col_client_id=self.config.col_client_id,
                col_target=col_target,
            )

        train, test, save_test_id = None, None, False
        if self.config.test_size == 'predefined':
            train, test = self.split_dataset_predefined(
                all_data=client_features,
                data_path=self.config.data_path,
                col_client_id=self.config.col_client_id,
                test_ids_path=self.config.output_test_ids_path,
            )
        elif float(self.config.test_size) > 0:
            # description
            spark.sparkContext.setLocalProperty('callSite.short', 'split_dataset')
            train, test = self.split_dataset(
                all_data=client_features,
                test_size=float(self.config.test_size),
                df_target=df_target,
                col_client_id=self.config.col_client_id,
                salt=self.config.salt,
            )
            save_test_id = True
        else:
            train = client_features

        # description
        spark.sparkContext.setLocalProperty('callSite.short', 'save_features')
        self.save_features(
            df_data=train,
            save_path=self.config.output_train_path,
        )

        if test is not None:
            self.save_features(
                df_data=test,
                save_path=self.config.output_test_path,
            )

        if save_test_id:
            self.save_test_ids(test)

        _duration = datetime.datetime.now() - _start
        logger.info(f'Data collected in {_duration.seconds} sec ({_duration})')

    def save_test_ids(self, df_test):
        test_ids = df_test.select(self.config.col_client_id).distinct().toPandas()
        test_ids.to_csv(self.config.output_test_ids_path, index=False)

    def load_target(self):
        df_target = self.load_source_data(self.config.target_files)
        return df_target

    def logging_config(self):
        if self.config.log_file is not None:
            handlers = [logging.StreamHandler(), logging.FileHandler(self.config.log_file, mode='w')]
        else:
            handlers = None
        logging.basicConfig(level=logging.INFO, format='%(funcName)-20s   : %(message)s', handlers=handlers)

    def load_transactions(self):
        spark = SparkSession.builder.getOrCreate()

        source_data = self.load_source_data(trx_files=self.config.trx_files)

        if len(self.config.dict) > 0:
            if len(self.config.dict) % 2 != 0:
                raise AttributeError('--dict should be in format (file_name col_id)*')
            for i in range(len(self.config.dict) // 2):
                # description
                spark.sparkContext.setLocalProperty('callSite.short', f'join with {i}th dict')
                source_data = self.join_dict(source_data, self.config.dict[2 * i], self.config.dict[2 * i + 1])

        # event_time mapping
        cols_event_time = self.config.cols_event_time
        if cols_event_time[0][0] == '#':
            if cols_event_time[0] == '#float':
                source_data = self._td_float(source_data, cols_event_time[1])
            elif cols_event_time[0] == '#datetime':
                source_data = self._td_datetime(source_data, cols_event_time[1])
            elif cols_event_time[0] == '#gender':
                source_data = self._td_gender(source_data, cols_event_time[1])
            else:
                raise NotImplementedError(f'Unknown type of data transformation: "{cols_event_time[0]}"')
        else:
            source_data = self._td_default(source_data, cols_event_time)

        return source_data


if __name__ == '__main__':
    DatasetConverter().run()



from ptls.data_load import padded_collate, TrxDataset
from ptls.nn.trx_encoder.trx_mean_encoder import TrxMeanEncoder
from ptls_tests.utils.data_generation import gen_trx_data


def test_mean_encoder():
    x, y = padded_collate(TrxDataset(gen_trx_data([4, 3, 2])))

    params = {
        "embeddings_noise": .1,
        'embeddings': {
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        'numeric_values': {'amount': 'log'}
    }

    te = TrxMeanEncoder(params)

    e = te(x)

    assert e.shape == (3, 33)



import torch
from torch.nn import functional as F


class VicregLoss(torch.nn.Module):
    """
    From https://github.com/facebookresearch/vicreg

    """
    def __init__(self, sim_coeff, std_coeff, cov_coeff):
        super(VicregLoss, self).__init__()

        self.sim_coeff = sim_coeff
        self.std_coeff = std_coeff
        self.cov_coeff = cov_coeff

    def forward(self, model_outputs, target):
        n = len(model_outputs)
        m = len(model_outputs[0])
        ix1 = torch.arange(0, n, 2, device=model_outputs.device)
        ix2 = torch.arange(1, n, 2, device=model_outputs.device)

        assert (target[ix1] == target[ix2]).all(), "Wrong embedding positions"

        x = model_outputs[ix1]
        y = model_outputs[ix2]

        # From https://github.com/facebookresearch/vicreg:

        repr_loss = F.mse_loss(x, y)

        x = x - x.mean(dim=0)
        y = y - y.mean(dim=0)

        std_x = torch.sqrt(x.var(dim=0) + 0.0001)
        std_y = torch.sqrt(y.var(dim=0) + 0.0001)
        std_loss = torch.mean(F.relu(1 - std_x)) / 2 +\
                   torch.mean(F.relu(1 - std_y)) / 2

        cov_x = (x.T @ x) / (n - 1)
        cov_y = (y.T @ y) / (n - 1)
        cov_loss = self.off_diagonal(cov_x).pow_(2).sum().div(m) +\
                   self.off_diagonal(cov_y).pow_(2).sum().div(m)

        loss = (self.sim_coeff * repr_loss +
                self.std_coeff * std_loss +
                self.cov_coeff * cov_loss)
        return loss

    @staticmethod
    def off_diagonal(x):
        n, m = x.shape
        assert n == m
        return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


import logging
import datetime
import numpy as np
from functools import reduce
from operator import iadd
from pyspark.sql import functions as F
from pyspark.sql import types as T
from pyspark.sql.window import Window
from itertools import chain
from typing import List, Dict, Union

from .base import DataPreprocessor, ColTransformer

from .base.col_category_transformer import ColCategoryTransformer
from .pyspark.category_identity_encoder import CategoryIdentityEncoder
from .pyspark.col_identity_transformer import ColIdentityEncoder
from .pyspark.event_time import DatetimeToTimestamp
from .pyspark.frequency_encoder import FrequencyEncoder
from .pyspark.user_group_transformer import UserGroupTransformer


logger = logging.getLogger(__name__)


class PysparkDataPreprocessor(DataPreprocessor):
    """Data preprocessor based on pyspark.sql.DataFrame

    During preprocessing it
        * transform `cols_event_time` column with date and time
        * encodes category columns `cols_category` into ints;
        * apply logarithm transformation to `cols_log_norm' columns;
        * (Optional) select the last `max_trx_count` transactions for each `col_id`;
        * groups flat data by `col_id`;
        * arranges data into list of dicts with features

    Parameters
    ----------
    col_id : str
        name of column with ids
    cols_event_time : str,
        name of column with time and date
    cols_category : list[str],s
        list of category columns
    cols_log_norm : list[str],
        list of columns to be logarithmed
    cols_identity : list[str],
        list of columns to be passed as is without any transformation
    cols_target: List[str],
        list of columns with target
    time_transformation: str. Default: 'default'.
        type of transformation to be applied to time column
    remove_long_trx: bool. Default: False.
        If True, select the last `max_trx_count` transactions for each `col_id`.
    max_trx_count: int. Default: 5000.
        used when `remove_long_trx`=True
    print_dataset_info : bool. Default: False.
        If True, print dataset stats during preprocessor fitting and data transformation
    """
    def __init__(self,
                 col_id: str,
                 col_event_time: Union[str, ColTransformer],
                 event_time_transformation: str = 'dt_to_timestamp',
                 cols_category: List[Union[str, ColCategoryTransformer]] = None,
                 category_transformation: str = 'frequency',
                 cols_numerical: List[str] = None,
                 cols_identity: List[str] = None,
                 cols_last_item: List[str] = None,
                 max_trx_count: int = None,
                 max_cat_num: Union[Dict[str, int], int] = 10000,
                 ):

        if cols_category is None:
            cols_category = []
        if cols_numerical is None:
            cols_numerical = []
        if cols_identity is None:
            cols_identity = []
        if cols_last_item is None:
            cols_last_item = []

        if type(col_event_time) is not str:
            ct_event_time = col_event_time  # use as is
        elif event_time_transformation == 'dt_to_timestamp':
            ct_event_time = DatetimeToTimestamp(col_name_original=col_event_time)
        elif event_time_transformation == 'none':
            ct_event_time = ColIdentityEncoder(
                col_name_original=col_event_time,
                col_name_target='event_time',
                is_drop_original_col=False,
            )
        else:
            raise AttributeError(f'incorrect event_time parameters combination: '
                                 f'`ct_event_time` = "{col_event_time}" '
                                 f'`event_time_transformation` = "{event_time_transformation}"')

        cts_category = []
        for col in cols_category:
            if type(col) is not str:
                cts_category.append(col)  # use as is
            elif category_transformation == 'frequency':
                if type(max_cat_num) is dict:
                    mc = max_cat_num.get(col)
                else:
                    mc = max_cat_num
                cts_category.append(FrequencyEncoder(col_name_original=col, max_cat_num=mc))
            elif category_transformation == 'none':
                cts_category.append(CategoryIdentityEncoder(col_name_original=col))
            else:
                raise AttributeError(f'incorrect category parameters combination: '
                                     f'`cols_category[i]` = "{col}" '
                                     f'`category_transformation` = "{category_transformation}"')

        cts_numerical = [ColIdentityEncoder(col_name_original=col) for col in cols_numerical]
        t_user_group = UserGroupTransformer(
            col_name_original=col_id, cols_last_item=cols_last_item, max_trx_count=max_trx_count)

        super().__init__(
            ct_event_time=ct_event_time,
            cts_category=cts_category,
            cts_numerical=cts_numerical,
            cols_identity=cols_identity,
            t_user_group=t_user_group,
        )

    @staticmethod
    def _td_default(df, cols_event_time):
        w = Window().orderBy(cols_event_time)
        tmp_df = df.select(cols_event_time).distinct()
        tmp_df = tmp_df.withColumn('event_time', F.row_number().over(w) - 1)
        df = df.join(tmp_df, on=cols_event_time)
        return df


    @staticmethod
    def _td_float(df, col_event_time):
        logger.info('To-float time transformation begins...')
        df = df.withColumn('event_time', F.col(col_event_time).astype('float'))
        logger.info('To-float time transformation ends')
        return df


    @staticmethod
    def _td_gender(df, col_event_time):
        """Gender-dataset-like transformation
        'd hh:mm:ss' -> float where integer part is day number and fractional part is seconds from day begin
        '1 00:00:00' -> 1.0
        '1 12:00:00' -> 1.5
        '1 01:00:00' -> 1 + 1 / 24
        '2 23:59:59' -> 1.99
        '432 12:00:00' -> 432.5   '000432 12:00:00'
        :param df:
        :param col_event_time:
        :return:
        """
        logger.info('Gender-dataset-like time transformation begins...')
        df = df.withColumn('_et_day', F.substring(F.lpad(F.col(col_event_time), 15, '0'), 1, 6).cast('float'))

        df = df.withColumn('_et_time', F.substring(F.lpad(F.col(col_event_time), 15, '0'), 8, 8))
        df = df.withColumn('_et_time', F.regexp_replace('_et_time', r'\:60$', ':59'))
        df = df.withColumn('_et_time', F.unix_timestamp('_et_time', 'HH:mm:ss') / (24 * 60 * 60))

        df = df.withColumn('event_time', F.col('_et_day') + F.col('_et_time'))
        df = df.drop('_et_day', '_et_time')
        logger.info('Gender-dataset-like time transformation ends')
        return df


    def _td_hours(self, df, col_event_time):
        logger.info('To hours time transformation begins...')
        df = df.withColumn('_dt', (F.col(col_event_time)).cast(dataType=T.TimestampType()))
        df = df.withColumn('event_time', ((F.col('_dt')).cast('float') - self.time_min) / 3600)
        df = df.drop('_dt')
        logger.info('To hours time transformation ends')
        return df


    def _reset(self):
        """Reset internal data-dependent state of the preprocessor, if necessary.
        __init__ parameters are not touched.
        """
        self.time_min = None
        self.remove_long_trx = False
        self.max_trx_count = 5000
        super()._reset()


    def pd_hist(self, df, name, bins=10):
        # logger.info('pd_hist begin')
        # logger.info(f'sf = {self.config.sample_fraction}')
        data = df.select(name)
        if self.config.sample_fraction is not None:
            data = data.sample(fraction=self.config.sample_fraction)
        data = data.toPandas()[name]

        if data.dtype.kind == 'f':
            round_len = 1 if data.max() > bins + 1 else 2
            bins = np.linspace(data.min(), data.max(), bins + 1).round(round_len)
        elif np.percentile(data, 99) - data.min() > bins - 1:
            bins = np.linspace(data.min(), np.percentile(data, 99), bins).astype(int).tolist() + [int(data.max() + 1)]
        else:
            bins = np.arange(data.min(), data.max() + 2, 1).astype(int)
        df = pd.cut(data, bins, right=False).rename(name)
        df = df.to_frame().assign(cnt=1).groupby(name)[['cnt']].sum()
        df['% of total'] = df['cnt'] / df['cnt'].sum()
        return df




import torch


class PairwiseMatrixSelector:
    """
    Returns matrix with one positive pair on first index and all negative pairs 
    for every possible pair. 
    """
    def get_pair_matrix(self, embeddings, labels):
        n = labels.size(0)

        uniq_num = len(torch.unique(labels))
        split_num = len(labels)//uniq_num

        # construct matrix x, such as x_ij == 1 <==> labels[i] == labels[j]
        x = labels.expand(n, n) - labels.expand(n, n).t()
        x.fill_diagonal_(1)

        indx = torch.where(x == 0)
        positive_pairs = torch.cat((indx[0].reshape(-1, 1), indx[1].reshape(-1, 1)), dim=1)
        positive_pairs = torch.unsqueeze(positive_pairs, 1)

        x.fill_diagonal_(0)
        indx = torch.where(x != 0)

        negative_pairs = torch.cat((indx[0].reshape(-1, 1), indx[1].reshape(-1, 1)), dim=1)
        negative_pairs = negative_pairs.reshape(-1, n - split_num, 2)
        negative_pairs = negative_pairs.repeat(1, split_num - 1, 1)
        negative_pairs = negative_pairs.reshape(-1, n - split_num, 2)

        pairs = torch.cat((positive_pairs, negative_pairs), dim=1)
        return embeddings[pairs]



import pyspark.sql
from pyspark.sql import functions as F

from ptls.preprocessing.base import ColTransformer
from ptls.preprocessing.pyspark.col_transformer import ColTransformerPysparkMixin


def dt_to_timestamp(col: str):
    return F.unix_timestamp(F.col(col).cast('timestamp'))


def timestamp_to_dt(col: str):
    return F.from_unixtime(col).cast('timestamp')


class DatetimeToTimestamp(ColTransformerPysparkMixin, ColTransformer):
    def __init__(self,
                 col_name_original: str = 'event_time',
                 is_drop_original_col: bool = True,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target='event_time',
            is_drop_original_col=is_drop_original_col,
        )

    def transform(self, x: pyspark.sql.DataFrame):
        x = x.withColumn(self.col_name_target, dt_to_timestamp(self.col_name_original))
        x = super().transform(x)
        return x



import logging
from copy import deepcopy

import pandas as pd
import pytorch_lightning as pl
import torch
import torchmetrics
from omegaconf import DictConfig

from ptls.data_load.padded_batch import PaddedBatch

logger = logging.getLogger(__name__)


class SequenceToTarget(pl.LightningModule):
    """LightningModule for supervised task or for inference

    For binary classification problem use settings like:
        head=torch.nn.Sequential(
            ...
            torch.nn.Linear(in_size, 1),
            torch.nn.Sigmoid(),
            torch.nn.Flatten(start_dim=0),
        ),
        loss=BCELoss(),
        metric_list=torchmetrics.AUROC(num_classes=2),

    For multiclass problem use settings like:
        head=torch.nn.Sequential(
            torch.nn.Linear(in_size, num_classes),
            torch.nn.LogSoftmax(dim=1),
        ),
        loss=torch.nn.NLLLoss(),
        metric_list=torchmetrics.Accuracy(compute_on_step=False),

    For regression problem use settings like:
        head=torch.nn.Sequential(
            torch.nn.Linear(in_size, 1),
            torch.nn.Flatten(start_dim=0),
        ),
        loss=torch.nn.MSELoss(),
        metric_list=torchmetrics.MeanSquaredError(),

    For `seq_encoder` inference:
        Just set `seq_encoder` and `head` (if needed), keep other parameters as None.
        Next use `torch.vstack(trainer.predict(sequence_to_target, dl_inference))`
        This call `seq_encoder.forward` which provide embeddings.

    Parameters
    ----------
    seq_encoder:
        Sequence encoder. May be pretrained or with random initialisation
    head:
        Head layers for your problem. May be simple or multilayer or None.
    loss:
        Your loss for specific problem.
    metric_list:
        One or dict of metrics for specific problem: {'metric_name1': metric1, 'metric_name2': metric2, ...}.
    optimizer_partial:
        optimizer init partial. Network parameters are missed.
    lr_scheduler_partial:
        scheduler init partial. Optimizer are missed.
    pretrained_lr:
        lr for seq_encoder. Can be one of:
        - 'freeze' - seq_encoder will be frozen
        - float - correspond seq_encoder lr, other network parameters have lr from optimizer_partial
        - None - all network parameters have lr from optimizer_partial
    train_update_n_steps:
        Interval of train metric update. Use it for large train dataset with heavy metrics.
        Update train metrics on each step when `train_update_n_steps is None`

    """

    def __init__(self,
                 seq_encoder: torch.nn.Module,
                 head: torch.nn.Module=None,
                 loss: torch.nn.Module=None,
                 metric_list: torchmetrics.Metric=None,
                 optimizer_partial=None,
                 lr_scheduler_partial=None,
                 pretrained_lr=None,
                 train_update_n_steps=None,
                 ):
        super().__init__()

        self.save_hyperparameters(ignore=[
            'seq_encoder', 'head', 'loss', 'metric_list', 'optimizer_partial', 'lr_scheduler_partial'])

        self.seq_encoder = seq_encoder
        self.head = head
        self.loss = loss

        if type(metric_list) is dict or type(metric_list) is DictConfig:
            metric_list = [(k, v) for k, v in metric_list.items()]
        else:
            if type(metric_list) is not list:
                metric_list = [metric_list]
            metric_list = [(m.__class__.__name__, m) for m in metric_list]

        self.train_metrics = torch.nn.ModuleDict([(name, deepcopy(mc)) for name, mc in metric_list])
        self.valid_metrics = torch.nn.ModuleDict([(name, deepcopy(mc)) for name, mc in metric_list])
        self.test_metrics = torch.nn.ModuleDict([(name, deepcopy(mc)) for name, mc in metric_list])

        self.optimizer_partial = optimizer_partial
        self.lr_scheduler_partial = lr_scheduler_partial

    def forward(self, x):
        x = self.seq_encoder(x)
        if self.head is not None:
            x = self.head(x)
        return x

    def training_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        loss = self.loss(y_h, y)
        self.log('loss', loss)
        if isinstance(x, PaddedBatch):
            self.log_dict(
                {"seq_len": x.seq_lens.float().mean(),
                 "y": y.unsqueeze(-1).sum(1).squeeze().float().mean(),
                }, prog_bar=True)
        train_update_n_steps = self.hparams.train_update_n_steps
        if train_update_n_steps is None or \
                train_update_n_steps is not None and self.global_step % train_update_n_steps == 0:
            for name, mf in self.train_metrics.items():
                mf(y_h, y)
        return loss

    def training_epoch_end(self, outputs):
        for name, mf in self.train_metrics.items():
            self.log(f'train_{name}', mf.compute(), prog_bar=False)
        for name, mf in self.train_metrics.items():
            mf.reset()

    def validation_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        self.log('val_loss', self.loss(y_h, y))
        for name, mf in self.valid_metrics.items():
            mf(y_h, y)

    def validation_epoch_end(self, outputs):
        for name, mf in self.valid_metrics.items():
            self.log(f'val_{name}', mf.compute(), prog_bar=True)
        for name, mf in self.valid_metrics.items():
            mf.reset()

    def test_step(self, batch, _):
        x, y = batch
        y_h = self(x)
        for name, mf in self.test_metrics.items():
            mf(y_h, y)

    def test_epoch_end(self, outputs):
        for name, mf in self.test_metrics.items():
            value = mf.compute().item()
            self.log(f'test_{name}', value, prog_bar=False)
        for name, mf in self.test_metrics.items():
            mf.reset()

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        if isinstance(batch, PaddedBatch):
            return self(batch)
        x, seq_id = batch
        pred = self.seq_encoder(x)
        x = x.drop_seq_features()
        x["seq_id"] = seq_id
        if self.head is not None:
            pred = self.head(pred)
            x["out"] = pred
        else:
            x["emb"] = pred
        return self.to_pandas(x)

    def configure_optimizers(self):
        if self.hparams.pretrained_lr is not None:
            if self.hparams.pretrained_lr == 'freeze':
                for p in self.seq_encoder.parameters():
                    p.requires_grad = False
                logger.info('Created optimizer with frozen encoder')
                parameters = self.parameters()
            else:
                parameters = [
                    {'params': self.seq_encoder.parameters(), 'lr': self.hparams.pretrained_lr},
                    {'params': self.head.parameters()},  # use predefined lr from `self.optimizer_partial`
                ]
                logger.info('Created optimizer with two lr groups')
        else:
            parameters = self.parameters()

        optimizer = self.optimizer_partial(parameters)
        scheduler = self.lr_scheduler_partial(optimizer)
        return [optimizer], [scheduler]

    @staticmethod
    def to_pandas(x):
        expand_cols = []
        scalar_features = {}

        for k, v in x.items():
            if type(v) is torch.Tensor:
                v = v.cpu().numpy()

            if len(v.shape) == 1:
                scalar_features[k] = v
            elif len(v.shape) == 2:
                expand_cols.append(k)
            else:
                scalar_features[k] = None

        dataframes = [pd.DataFrame(scalar_features)]
        for col in expand_cols:
            v = x[col].cpu().numpy()
            dataframes.append(pd.DataFrame(v, columns=[f'{col}_{i:04d}' for i in range(v.shape[1])]))

        return pd.concat(dataframes, axis=1)



from torch import nn as nn
from torch.autograd import Function


class Binarization(Function):
    @staticmethod
    def forward(self, x):
        q = (x > 0).float()
        return 2*q - 1

    @staticmethod
    def backward(self, grad_outputs):
        return grad_outputs


binary = Binarization.apply


class BinarizationLayer(nn.Module):
    def __init__(self, hs_from, hs_to):
        super(BinarizationLayer, self).__init__()
        self.linear = nn.Linear(hs_from, hs_to, bias=False)

    def forward(self, x):
        return binary(self.linear(x))


import random
from functools import reduce
from operator import iadd

import torch

from ptls.data_load.augmentations.sequence_pair_augmentation import sequence_pair_augmentation
from ptls.data_load.utils import collate_feature_dict
from ptls.frames.coles import ColesDataset


class NspDataset(ColesDataset):
    def get_splits(self, feature_arrays):
        return [sequence_pair_augmentation(item)
                for item in super().get_splits(feature_arrays)]

    @staticmethod
    def collate_fn(batch):
        batch = reduce(iadd, batch)
        lefts = [left for left, _ in batch] * 2

        rights = [right for _, right in batch]
        rights_ = rights[:]
        random.shuffle(rights_)
        rights += rights_

        targets = torch.cat([
            torch.ones(len(batch), dtype=torch.int64),
            torch.zeros(len(batch), dtype=torch.int64),
        ])

        return (
            (
                collate_feature_dict(lefts),
                collate_feature_dict(rights)
            ),
            targets.long(),
        )


class NspIterableDataset(NspDataset, torch.utils.data.IterableDataset):
    pass



from ptls.data_load.augmentations.seq_len_limit import SeqLenLimit
import numpy as np


def test_tail_no_limit():
    i_filter = SeqLenLimit(max_seq_len=10)
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.arange(8))


def test_tail_with_limit():
    i_filter = SeqLenLimit(max_seq_len=4)
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.array([4, 5, 6, 7]))


def test_head_no_limit():
    i_filter = SeqLenLimit(max_seq_len=10, strategy='head')
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.arange(8))


def test_head_with_limit():
    i_filter = SeqLenLimit(max_seq_len=4, strategy='head')
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.array([0, 1, 2, 3]))


def test_random_no_limit():
    i_filter = SeqLenLimit(max_seq_len=10, strategy='random')
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.arange(8))


def test_random_with_limit():
    i_filter = SeqLenLimit(max_seq_len=2, strategy='random')
    data = {'mcc': np.arange(8)}
    data = i_filter(data)
    assert len(data['mcc']) == 2
    assert data['mcc'][0] + 1 == data['mcc'][1]


def test_tail_with_limit_and_target():
    i_filter = SeqLenLimit(max_seq_len=4)
    data = {'mcc': np.arange(8), 'target': 0}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.array([4, 5, 6, 7]))



import numpy as np

from ptls.data_load.augmentations.random_slice import RandomSlice


def test_usage():
    i_filter = RandomSlice(10, 20)
    data = {'mcc': np.arange(100)}
    data = i_filter(data)
    assert 10 <= len(data['mcc']) <= 20
    assert (np.diff(data['mcc']) == 1).all()


def test_len_1():
    i_filter = RandomSlice(10, 20)
    min_len, max_len = i_filter.get_min_max(100)
    assert (min_len, max_len) == (10, 20)


def test_len_2():
    i_filter = RandomSlice(10, 20)
    min_len, max_len = i_filter.get_min_max(15)
    assert (min_len, max_len) == (10, 15)


def test_len_3():
    i_filter = RandomSlice(10, 20)
    min_len, max_len = i_filter.get_min_max(5)
    assert (min_len, max_len) == (5, 5)


def test_len_3_full():
    i_filter = RandomSlice(10, 20)
    data = {'mcc': np.arange(5)}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.arange(5))


def test_with_target():
    i_filter = RandomSlice(10, 20)
    data = {'mcc': np.arange(5), 'target': 0}
    data = i_filter(data)
    np.testing.assert_equal(data['mcc'], np.arange(5))


def test_len_4():
    i_filter = RandomSlice(10, 20, rate_for_min=0.8)
    min_len, max_len = i_filter.get_min_max(100)
    assert (min_len, max_len) == (10, 20)


def test_len_5():
    i_filter = RandomSlice(10, 20, rate_for_min=0.8)
    min_len, max_len = i_filter.get_min_max(15)
    assert (min_len, max_len) == (10, 15)


def test_len_6():
    i_filter = RandomSlice(10, 20, rate_for_min=0.8)
    min_len, max_len = i_filter.get_min_max(5)
    assert (min_len, max_len) == (4, 5)


def test_len_7():
    i_filter = RandomSlice(10, 20, rate_for_min=0.01)
    min_len, max_len = i_filter.get_min_max(5)
    assert (min_len, max_len) == (1, 5)



import torch
import warnings
from torch import nn as nn


class CentroidLoss(nn.Module):
    """Centroid loss

    Class centroids are calculated over batch

    Parameters
    ----------
    class_num
        set the number of classes if you know limit (e.g. 2 for binary classifications)
        set None for metric learning task with unknown class number
    centroid_margin
        l2 distance between the class centers, closer than which the loss will be calculated.
        Class centers tend to be further than `centroid_margin`.
    """

    def __init__(self, class_num, centroid_margin=1.4):
        super().__init__()
        self.class_num = class_num
        if self.class_num is not None:
            self.register_buffer('l_targets_eye', torch.eye(class_num))
        self.centroid_margin = centroid_margin
        self.eps = 1e-6

    def forward(self, embeddings, target):
        if self.class_num is None:
            class_num = target.max() + 1
            l_targets_eye = torch.eye(class_num, device=target.device)
        else:
            l_targets_eye = self.l_targets_eye

        l_targets_ohe = l_targets_eye[target]  # B, class
        class_centers = embeddings.unsqueeze(1) * l_targets_ohe.unsqueeze(2) # B, class, H
        class_centers = class_centers.sum(dim=0).div(l_targets_ohe.sum(dim=0).unsqueeze(1) + self.eps)  # class, H

        cc_pairs = (class_centers.unsqueeze(1) - class_centers.unsqueeze(0)).pow(2).sum(dim=2)
        cc_pairs = torch.relu(self.centroid_margin - (cc_pairs + self.eps).pow(0.5)).pow(2)
        cc_pairs = torch.triu(cc_pairs, diagonal=1).sum() / cc_pairs.size(0) * (cc_pairs.size(0) - 1) / 2

        distances = embeddings - class_centers[target]
        l2_loss = distances.pow(2).sum(dim=1)  # B,

        return l2_loss.mean() + cc_pairs


class CentroidSoftmaxLoss(nn.Module):
    """Centroid Softmax loss

    Class centroids are calculated over batch

    Parameters
    ----------
    class_num
        set the number of classes if you know limit (e.g. 2 for binary classifications)
        set None for metric learning task with unknown class number
    temperature:
        temperature for softmax logits for scaling l2 distance on unit sphere
    """

    def __init__(self, class_num, temperature=10.0):
        super().__init__()
        self.class_num = class_num
        if self.class_num is not None:
            self.register_buffer('l_targets_eye', torch.eye(class_num))
        self.eps = 1e-6
        self.temperature = temperature

    def forward(self, embeddings, target):
        if self.class_num is None:
            class_num = target.max() + 1
            l_targets_eye = torch.eye(class_num, device=target.device)
        else:
            l_targets_eye = self.l_targets_eye

        l_targets_ohe = l_targets_eye[target]  # B, class
        class_centers = embeddings.unsqueeze(1) * l_targets_ohe.unsqueeze(2) # B, class, H
        class_centers = class_centers.sum(dim=0).div(l_targets_ohe.sum(dim=0).unsqueeze(1) + self.eps)  # class, H

        distances = embeddings.unsqueeze(1) - class_centers.unsqueeze(0)  # B, class, H
        l2_loss = distances.pow(2).sum(dim=2)  # B, class

        l2_loss = torch.softmax(-l2_loss * self.temperature, dim=1)
        l2_loss = l2_loss[torch.arange(l2_loss.size(0), device=embeddings.device), target]
        l2_loss = -torch.log(l2_loss)

        return l2_loss.mean()


class CentroidSoftmaxMemoryLoss(nn.Module):
    """Centroid Softmax Memory loss

    Class centroids are calculated over batch and saved in memory as running average

    Parameters
    ----------
    class_num
        set the number of classes if you know limit (e.g. 2 for binary classifications)
        set None for metric learning task with unknown class number
    temperature:
        temperature for softmax logits for scaling l2 distance on unit sphere
    alpha:
        rate of history keep for running average
    """

    def __init__(self, class_num, hidden_size, temperature=10.0, alpha=0.99):
        super().__init__()
        assert class_num is not None
        self.class_num = class_num
        self.register_buffer('l_targets_eye', torch.eye(class_num))
        self.eps = 1e-6
        self.temperature = temperature
        self.alpha = alpha

        self.register_buffer('class_centers', torch.zeros(class_num, hidden_size, dtype=torch.float))
        self.is_empty_class_centers = True

    def forward(self, embeddings, target):
        l_targets_eye = self.l_targets_eye

        l_targets_ohe = l_targets_eye[target]  # B, class
        class_centers = embeddings.unsqueeze(1) * l_targets_ohe.unsqueeze(2) # B, class, H
        class_centers = class_centers.sum(dim=0).div(l_targets_ohe.sum(dim=0).unsqueeze(1) + self.eps)  # class, H

        class_centers = class_centers.detach()
        if not self.is_empty_class_centers:
            self.class_centers = self.class_centers * self.alpha + class_centers * (1 - self.alpha)
        else:
            self.class_centers = class_centers
            self.is_empty_class_centers = False

        distances = embeddings.unsqueeze(1) - self.class_centers.unsqueeze(0)  # B, class, H
        l2_loss = distances.pow(2).sum(dim=2)  # B, class

        l2_loss = torch.softmax(-l2_loss * self.temperature, dim=1)
        l2_loss = l2_loss[torch.arange(l2_loss.size(0), device=embeddings.device), target]
        l2_loss = -torch.log(l2_loss)

        return l2_loss.mean()



import torch


class QuerySoftmaxLoss(torch.nn.Module):
    def __init__(self, temperature: float = 1.0, reduce: bool = True):
        """

        Parameters
        ----------
        temperature:
            softmax(logits * temperature)
        reduce:
            if `reduce` then `loss.mean()` returned. Else loss by elements returned
        """

        super().__init__()
        self.temperature = temperature
        self.reduce = reduce

    def forward(self, anchor, pos, neg):
        logits = self.get_logits(anchor, pos, neg)
        probas = torch.softmax(logits, dim=1)
        loss = -torch.log(probas[:, 0])
        if self.reduce:
            return loss.mean()
        return loss

    def get_logits(self, anchor, pos, neg):
        all_counterparty = torch.cat([pos, neg], dim=1)
        logits = (anchor * all_counterparty).sum(dim=2) * self.temperature
        return logits


from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class FeatureRename(IterableProcessingDataset):
    def __init__(self, feature_names: dict):
        """Rename features in dict

        Params:
            feature_names: keys are original names, values are target names
        """
        super().__init__()

        self._feature_names = feature_names

    def process(self, features):
        return {self._feature_names.get(k, k): v for k, v in features.items()}



from ptls.data_load.iterable_processing.target_move import TargetMove
import numpy as np
import torch


def get_data():
    return [
        {
            'client_id': 1,
            'trans_date': np.array([0., 1., 2., 5., 8., 9., 11., 15.]),
            'small_group': np.array([0, 1, 2, 5, 8, 9, 11, 15]),
            'target': np.array(1)
        }
    ]


def test_target_move_filter():
    target_col = 'target'
    ifilter = TargetMove(target_col)
    data_before = get_data()
    data_after = list(ifilter(data_before))[0]
    data_before = data_before[0]
    assert isinstance(data_after, tuple)
    assert len(data_after) == 2
    target_val = data_before[target_col].item()
    assert target_val == data_after[1]
    for k, v in data_before.items():
        if isinstance(v, np.ndarray):
            assert np.array_equal(v, data_after[0][k])
        elif isinstance(v, torch.Tensor):
            assert torch.equal(v, data_after[0][k])
        else:
            assert v == data_after[0][k]



from itertools import combinations

import numpy as np
import torch

from ptls.frames.coles.sampling_strategies.triplet_selector import TripletSelector


class AllTripletSelector(TripletSelector):
    """
    Returns all possible triplets
    May be impractical in most cases
    """

    def __init__(self):
        super(AllTripletSelector, self).__init__()

    def get_triplets(self, embeddings, labels):
        np_labels = labels.cpu().data.numpy()
        triplets = []
        for label in set(np_labels):
            label_mask = (np_labels == label)
            label_indices = np.where(label_mask)[0]
            if len(label_indices) < 2:
                continue
            negative_indices = np.where(np.logical_not(label_mask))[0]
            anchor_positives = list(combinations(label_indices, 2))  # All anchor-positive pairs

            # Add all negatives for all positive pairs
            temp_triplets = [[anchor_positive[0], anchor_positive[1], neg_ind] for anchor_positive in anchor_positives
                             for neg_ind in negative_indices]
            triplets += temp_triplets

        return torch.LongTensor(np.array(triplets)).to(labels.device)


import warnings

import pyspark
import pyspark.sql.functions as F

from ptls.preprocessing.base.col_category_transformer import ColCategoryTransformer
from ptls.preprocessing.pyspark.col_transformer import ColTransformerPysparkMixin


class CategoryIdentityEncoder(ColTransformerPysparkMixin, ColCategoryTransformer):
    """Keep encoding from original category column

    Let's `col_name_original` value_counts looks like this:
    cat value: records counts in dataset
          4:  100
          2:  50
          1:  10
          6:  1

    Mapping will use this order to enumerate embedding indexes for category values:
    cat value: embedding id
    <padding token>: 0
                1: 1
                2: 2
                4: 4
                6: 6
     <other values>: 6

    `dictionary_size` will be 7

    Note:
       - expect integer values in original column which are mentioned as embedding indexes
       - 0 index is reserved for padding value
       - negative indexes aren't allowed
       - there are no <other values>. Input and output are identical.

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.

    """
    def __init__(self,
                 col_name_original: str,
                 col_name_target: str = None,
                 is_drop_original_col: bool = True,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target=col_name_target,
            is_drop_original_col=is_drop_original_col,
        )

        self.min_fit_index = None
        self.max_fit_index = None

    def get_column(self, x):
        return x.withColumn(self.col_name_target, F.col(self.col_name_original).cast('int'))

    def fit(self, x: pyspark.sql.DataFrame):
        super().fit(x)
        x = self.get_column(x)
        agg = x.groupby().agg(
            F.min(self.col_name_target).alias('min'),
            F.max(self.col_name_target).alias('max'),
        ).collect()
        self.min_fit_index, self.max_fit_index = agg[0]['min'], agg[0]['max']
        if self.min_fit_index < 0:
            raise AttributeError(f'Negative values found in {self.col_name_original}')
        if self.min_fit_index == 0:
            warnings.warn(f'0 values fount in {self.col_name_original}. 0 is a padding index', UserWarning)
        return self

    @property
    def dictionary_size(self):
        return self.max_fit_index + 1

    def transform(self, x: pyspark.sql.DataFrame):
        x = self.get_column(x)
        agg = x.groupby().agg(
            F.min(self.col_name_target).alias('min'),
            F.max(self.col_name_target).alias('max'),
        ).collect()
        min_index, max_index = agg[0]['min'], agg[0]['max']
        if min_index < self.min_fit_index:
            warnings.warn(f'Not fitted values. min_index({min_index}) < min_fit_index({self.min_fit_index})',
                          UserWarning)
        if max_index > self.max_fit_index:
            warnings.warn(f'Not fitted values. max_index({max_index}) < max_fit_index({self.max_fit_index})',
                          UserWarning)

        x = super().transform(x)
        return x



import pandas as pd
import pytorch_lightning as pl
import torch

from ptls.data_load.padded_batch import PaddedBatch


class InferenceModule(pl.LightningModule):
    def __init__(self, model, pandas_output=True, drop_seq_features=True, model_out_name='out'):
        super().__init__()

        self.model = model
        self.pandas_output = pandas_output
        self.drop_seq_features = drop_seq_features
        self.model_out_name = model_out_name

    def forward(self, x: PaddedBatch):
        out = self.model(x)
        if self.drop_seq_features:
            x = x.drop_seq_features()
            x[self.model_out_name] = out
        else:
            x.payload[self.model_out_name] = out
        if self.pandas_output:
            return self.to_pandas(x)
        return x

    @staticmethod
    def to_pandas(x):
        expand_cols = []
        scalar_features = {}

        for k, v in x.items():
            if type(v) is torch.Tensor:
                v = v.cpu().numpy()

            if len(v.shape) == 1:
                scalar_features[k] = v
            elif len(v.shape) == 2:
                expand_cols.append(k)
            else:
                scalar_features[k] = None

        dataframes = [pd.DataFrame(scalar_features)]
        for col in expand_cols:
            v = x[col].cpu().numpy()
            dataframes.append(pd.DataFrame(v, columns=[f'{col}_{i:04d}' for i in range(v.shape[1])]))

        return pd.concat(dataframes, axis=1)



from functools import partial

import omegaconf
import pytorch_lightning as pl
import torch
import hydra
import torchmetrics
from pyhocon import ConfigFactory

from ptls.frames.supervised import SequenceToTarget
from ptls.loss import BCELoss
from ptls.nn import PBLinear, RnnSeqEncoder, TransformerSeqEncoder, TrxEncoder
from ptls_tests.test_data_load import RandomEventData


def get_rnn_params():
    return dict(
        seq_encoder=RnnSeqEncoder(
            trx_encoder=TrxEncoder(
                norm_embeddings=False,
                embeddings_noise=0.1,
                embeddings={
                    'mcc_code': {'in': 21, 'out': 3},
                    'trans_type': {'in': 11, 'out': 2},
                },
                numeric_values={'amount': 'log'},
            ),
            hidden_size=16,
            type='gru',
        ),
        optimizer_partial=partial(torch.optim.Adam, lr=0.004),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=10, gamma=0.8),
    )


def tst_params_data():
    params = {
        "train": {
            "batch_size": 32,
            "num_workers": 1,
            "trx_dropout": .1,
            "max_seq_len": 30
        },
        "valid": {
            "batch_size": 32,
            "num_workers": 1,
            "max_seq_len": 30
        },
    }
    return ConfigFactory.from_dict(params)


def tst_params_transf():
    params = {
        'model_type': 'transf',
        'transf': {
            'n_heads': 2,
            'dim_hidden': 16,
            'dropout': .1,
            'n_layers': 2,
            'max_seq_len': 200,
            'use_after_mask': True,
            'use_positional_encoding': True,
            'sum_output': True,
            'input_size': 16,
            'train_starter': True,
            'shared_layers': False,
            'use_src_key_padding_mask': False
        },
        "head": {
            "explicit_lengths": False,
        }
    }

    return ConfigFactory.from_dict(params).with_fallback(tst_params_data())


def test_train_loop_rnn_binary_classification():
    model = SequenceToTarget(
        head=torch.nn.Sequential(
            torch.nn.Linear(16, 1),
            torch.nn.Sigmoid(),
            torch.nn.Flatten(start_dim=0),
        ),
        loss=BCELoss(),
        metric_list=torchmetrics.AUROC(num_classes=2),
        **get_rnn_params(),
    )
    dl = RandomEventData(tst_params_data(), target_type='bin_cls')
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)
    print(trainer.logged_metrics)


def test_train_loop_rnn_milti_classification():
    model = SequenceToTarget(
        head=torch.nn.Sequential(
            torch.nn.Linear(16, 4),
            torch.nn.LogSoftmax(dim=1),
        ),
        loss=torch.nn.NLLLoss(),
        metric_list={
            'auroc': torchmetrics.AUROC(num_classes=4),
            'accuracy': torchmetrics.Accuracy(),
        },
        **get_rnn_params(),
    )
    dl = RandomEventData(tst_params_data(), target_type='multi_cls')
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)
    print(trainer.logged_metrics)


def test_train_loop_rnn_regression():
    model = SequenceToTarget(
        head=torch.nn.Sequential(
            torch.nn.Linear(16, 1),
            torch.nn.Flatten(start_dim=0),
        ),
        loss=torch.nn.MSELoss(),
        metric_list=torchmetrics.MeanSquaredError(compute_on_step=False, squared=False),
        **get_rnn_params(),
    )
    dl = RandomEventData(tst_params_data(), target_type='regression')
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)
    print(trainer.logged_metrics)


def test_train_loop_transf():
    trx_encoder = TrxEncoder(
        norm_embeddings=False,
        embeddings_noise=0.1,
        embeddings={
            'mcc_code': {'in': 21, 'out': 3},
            'trans_type': {'in': 11, 'out': 2},
        },
        numeric_values={'amount': 'log'},
    )
    seq_encoder = TransformerSeqEncoder(
        trx_encoder=torch.nn.Sequential(trx_encoder, PBLinear(trx_encoder.output_size, 20)),
        input_size=20,
        n_heads=1,
        n_layers=1,
    )
    model = SequenceToTarget(
        seq_encoder=seq_encoder,
        head=torch.nn.Sequential(
            torch.nn.Linear(seq_encoder.embedding_size, 1),
            torch.nn.Sigmoid(),
            torch.nn.Flatten(start_dim=0),
        ),
        loss=BCELoss(),
        metric_list=torchmetrics.AUROC(num_classes=2),
        optimizer_partial=partial(torch.optim.Adam, lr=0.004),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=10, gamma=0.8),
    )
    dl = RandomEventData(tst_params_data())
    trainer = pl.Trainer(max_epochs=1, logger=None, checkpoint_callback=False)
    trainer.fit(model, dl)

# SequenceToTarget.metric_list
def test_seq_to_target_metric_list_single_metric():
    model = SequenceToTarget(metric_list=torchmetrics.Accuracy(), seq_encoder=None)
    metric_name = next(iter(model.valid_metrics.keys()))
    assert metric_name == 'Accuracy'


def test_seq_to_target_metric_list_list_with_metric():
    model = SequenceToTarget(metric_list=[
        torchmetrics.Accuracy(),
        torchmetrics.AUROC(num_classes=2),
    ], seq_encoder=None)
    assert 'Accuracy' in model.valid_metrics
    assert 'AUROC' in model.valid_metrics


def test_seq_to_target_metric_list_dict_with_single_metric():
    model = SequenceToTarget(metric_list={
        'acc': torchmetrics.Accuracy(),
    }, seq_encoder=None)
    metric_name = next(iter(model.valid_metrics.keys()))
    assert metric_name == 'acc'


def test_seq_to_target_metric_list_dict_with_metric():
    model = SequenceToTarget(metric_list={
        'acc': torchmetrics.Accuracy(),
        'auroc': torchmetrics.AUROC(num_classes=2),
    }, seq_encoder=None)
    assert 'acc' in model.valid_metrics
    assert 'auroc' in model.valid_metrics


def test_seq_to_target_metric_list_dict_config_with_metric():
    conf = omegaconf.OmegaConf.create("""
        auroc:
            _target_: torchmetrics.AUROC
            num_classes: 2
        acc:
            _target_: torchmetrics.Accuracy
    """)
    model = SequenceToTarget(metric_list=hydra.utils.instantiate(conf), seq_encoder=None)
    assert 'acc' in model.valid_metrics
    assert 'auroc' in model.valid_metrics



import pytest
import numpy as np

from ptls.data_load.iterable_processing.id_filter import IdFilter


def get_data(id_type):
    return [{'client_id': id_type(i)} for i in range(1, 10)]


def test_int():
    i_filter = IdFilter('client_id', [1, 5, 9])
    data = i_filter(get_data(int))
    data = [x['client_id'] for x in data]
    assert data == [1, 5, 9]


def test_np():
    i_filter = IdFilter('client_id', np.array([1, 5, 9]).astype(np.int16))
    data = i_filter(get_data(np.int16))
    data = [x['client_id'] for x in data]
    assert data == [1, 5, 9]


def test_type_mismatch_int_str():
    i_filter = IdFilter('client_id', [1, 5, 9])
    data = i_filter(get_data(str))
    with pytest.raises(TypeError):
        _ = [x['client_id'] for x in data]



from ptls.data_load.iterable_processing.feature_rename import FeatureRename


def test_feature_rename():
    x = {
        'mcc': [1, 2 ,3],
        'score': 1,
    }
    i_filter = FeatureRename({'score': 'target'})
    data = next(i_filter([x]))
    assert 'mcc' in data
    assert 'target' in data
    assert 'score' not in data



import numpy as np

from ptls.size_reduction import embeds_quantiles, pca_reduction


def test_embeds_quantiles():
    array = np.array([
        [0.1, 0.6, -0.4, 0.9],
        [0.2, 0.7, -0.3, 0.8],
        [0.3, 0.8, -0.2, 0.7],
        [0.4, 0.9, -0.1, 0.6],
        [0.1, 0.6, -0.4, 0.9],
        [0.2, 0.7, -0.3, 0.8],
        [0.3, 0.8, -0.2, 0.7],
        [0.4, 0.9, -0.1, 0.6]
    ])

    out = np.array([
        (0, 0, 0, 3),
        (1, 1, 1, 2),
        (2, 2, 2, 1),
        (3, 3, 3, 0),
        (0, 0, 0, 3),
        (1, 1, 1, 2),
        (2, 2, 2, 1),
        (3, 3, 3, 0),
    ])

    q_embeds, bins = embeds_quantiles(array, q_count=4, num_workers=0)
    assert np.array_equal(q_embeds, out)

    q_embeds = embeds_quantiles(array, bins=bins, num_workers=0)
    assert np.array_equal(q_embeds, out)

    q_embeds, bins = embeds_quantiles(array, q_count=4, num_workers=8)
    assert np.array_equal(q_embeds, out)

    q_embeds = embeds_quantiles(array, bins=bins, num_workers=8)
    assert np.array_equal(q_embeds, out)


def test_pca_reduction():
    array = np.array([
        [0.1,  0.6, -0.4,  0.9],
        [0.2,  0.7, -0.3,  0.8],
        [0.3,  0.8, -0.2,  0.7],
        [0.4,  0.9, -0.1,  0.6],
        [0.1,  0.6, -0.4,  0.9],
        [0.2,  0.7, -0.3,  0.8],
        [0.3,  0.8, -0.2,  0.7],
        [0.4,  0.9, -0.1,  0.6]
    ])

    out = np.array([[-0.3, -0.1,  0.1,  0.3, -0.3, -0.1,  0.1,  0.3]]).T

    reduced_array, pca_list = pca_reduction(array, sub_dim=4, n_components=1, num_workers=0)
    assert np.allclose(reduced_array, out, rtol=1e-07)

    reduced_array = pca_reduction(array, sub_dim=4, pca_list=pca_list, num_workers=0)
    assert np.allclose(reduced_array, out, rtol=1e-07)

    reduced_array, pca_list = pca_reduction(array, sub_dim=4, n_components=1, num_workers=8)
    assert np.allclose(reduced_array, out, rtol=1e-07)

    reduced_array = pca_reduction(array, sub_dim=4, pca_list=pca_list, num_workers=8)
    assert np.allclose(reduced_array, out, rtol=1e-07)



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class FeatureTypeCast(IterableProcessingDataset):
    def __init__(self, feature_types):
        """

        Args:
            feature_names: feature name for keep
        """
        super().__init__()

        self._feature_types = feature_types

    def process(self, features):
        return {k: self._feature_types.get(k, lambda x: x)(v)
                for k, v in features.items()}



from .contrastive_loss import ContrastiveLoss
from .margin_loss import MarginLoss
from .binomial_deviance_loss import BinomialDevianceLoss
from .barlow_twins_loss import BarlowTwinsLoss
from .vicreg_loss import VicregLoss

from .triplet_loss import TripletLoss

from .histogram_loss import HistogramLoss
from .centroid_loss import CentroidLoss, CentroidSoftmaxLoss, CentroidSoftmaxMemoryLoss

from .softmax_loss import SoftmaxLoss
from .softmax_pairwise_loss import SoftmaxPairwiseLoss



import torch

from ptls.data_load import PaddedBatch


class PBFeatureExtract(torch.nn.Module):
    def __init__(self, feature_col_name, as_padded_batch):
        super().__init__()

        self.feature_col_name = feature_col_name
        self.as_padded_batch = as_padded_batch

    def forward(self, x: PaddedBatch):
        feature = x.payload[self.feature_col_name]
        if self.as_padded_batch:
            return PaddedBatch(feature, x.seq_lens)
        return feature



from sklearn.base import BaseEstimator, TransformerMixin


class ColTransformer(BaseEstimator, TransformerMixin):
    """Base class for dataframe column transformer

    Check if columns exists.

    May rename or copy original column

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.
    """
    def __init__(self,
                 col_name_original: str,
                 col_name_target: str = None,
                 is_drop_original_col: bool = True,
                 ):
        super().__init__()
        self.col_name_original = col_name_original
        self.col_name_target = col_name_target if col_name_target is not None else col_name_original
        self.is_drop_original_col = is_drop_original_col

    def check_is_col_exists(self, x):
        raise NotImplementedError()

    def drop_original_col(self, x):
        raise NotImplementedError()

    def fit(self, x):
        self.check_is_col_exists(x)
        return self

    def transform(self, x):
        x = self.drop_original_col(x)
        return x



from .datasets.mlm_dataset import MlmDataset, MlmIterableDataset, MLMNSPDataset
from .datasets.mlm_indexed_dataset import MlmIndexedDataset, MLMNSPIndexedDataset
from .datasets.rtd_dataset import RtdDataset, RtdIterableDataset
from .datasets.sop_dataset import SopDataset, SopIterableDataset
from .datasets.nsp_dataset import NspDataset, NspIterableDataset

from .modules.mlm_module import MLMPretrainModule
from .modules.mlm_nsp_module import MLMNSPModule
from .modules.rtd_module import RtdModule
from .modules.sop_nsp_module import SopNspModule



import torch
import torch.nn as nn

class TabFormerFeatureEncoder(nn.Module):
    """TabFormerFeatureEncoder: encodes input batch of shape (B, T, F, E),
           where:
               B - batch size,
               T - sequence length,
               F - number of features
               E - embedding dimension for each feature
       and returns output batch of same shape.

       Encoding is performed as in [Tabular Transformers for Modeling Multivariate Time Series](https://arxiv.org/abs/2011.01843)

       Parameters
       ----------
       n_cols: number of features to encode,
       emb_dim: feature embedding dimension,
       n_heads: number of heads in transformer,
       n_layers: number of layers in transformer,
       out_hidden: out hidden dimension for each feature
    """

    def __init__(self, n_cols: int,
                       emb_dim: int,
                       transf_feedforward_dim: int = 64,
                       n_heads: int = 8,
                       n_layers: int = 1,
                       out_hidden: int = None):
        super().__init__()

        out_hidden = out_hidden if out_hidden else emb_dim * n_cols
        encoder_layer = nn.TransformerEncoderLayer(d_model=emb_dim, nhead=n_heads, dim_feedforward=transf_feedforward_dim, batch_first=True, activation='gelu')
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=n_layers)
        self.lin_proj = nn.Linear(emb_dim * n_cols, out_hidden)

    def forward(self, input_embeds):
        embeds_shape = list(input_embeds.size())
        input_embeds = input_embeds.view([-1] + embeds_shape[-2:])  # (B, T, NUM_F, H) -> (B*T, NUM_F, H)
        out_embeds = self.transformer_encoder(input_embeds)
        out_embeds = out_embeds.contiguous().view(embeds_shape[0:2]+[-1])  # (B*T, NUM_F, H) -> (B, T, NUM_F*H)
        out_embeds = self.lin_proj(out_embeds)  # (B, T, H1) -> (B, T, H2)
        return out_embeds




import logging

import hydra
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.multiprocessing
from omegaconf import DictConfig
from torch.utils.data.dataloader import DataLoader

from ptls.data_load.utils import collate_feature_dict
from ptls.frames.inference_module import InferenceModule

logger = logging.getLogger(__name__)


def save_scores(df_scores, output_conf):
    # output
    output_name = output_conf.path
    output_format = output_conf.format
    if output_format not in ('pickle', 'csv', 'parquet'):
        logger.warning(f'Format "{output_format}" is not supported. Used default "pickle"')
        output_format = 'pickle'

    output_path = f'{output_name}.{output_format}'

    if output_format == 'pickle':
        df_scores.to_pickle(output_path)
    elif output_format == 'csv':
        df_scores.to_csv(output_path, sep=',', header=True, index=False)
    elif output_format == 'parquet':
        df_scores.to_parquet(output_path)
    else:
        raise AssertionError('Never happens')
    logger.info(f'{len(df_scores)} records saved to: "{output_path}"')


@hydra.main(version_base='1.2', config_path=None)
def main(conf: DictConfig):
    if 'seed_everything' in conf:
        pl.seed_everything(conf.seed_everything)

    seq_encoder = hydra.utils.instantiate(conf.inference.seq_encoder)
    if type(seq_encoder) is DictConfig and seq_encoder.get('load_from_checkpoint', False):
        pl_module = hydra.utils.instantiate(conf.pl_module)
        pl_module.load_state_dict(torch.load(seq_encoder['f'])['state_dict'])
        seq_encoder = pl_module.seq_encoder
    model = InferenceModule(
        model=seq_encoder,
        pandas_output=True, model_out_name='emb',
    )
    model.model.is_reduce_sequence = True

    dataset_inference = hydra.utils.instantiate(conf.inference.dataset)
    inference_dl = DataLoader(
        dataset=dataset_inference,
        collate_fn=collate_feature_dict,
        shuffle=False,
        num_workers=conf.inference.get('num_workers', 0),
        batch_size=conf.inference.get('batch_size', 128),
    )

    gpus = 1 if torch.cuda.is_available() else 0
    gpus = conf.inference.get('gpus', gpus)
    df_scores = pl.Trainer(gpus=gpus, max_epochs=-1).predict(model, inference_dl)
    df_scores = pd.concat(df_scores, axis=0)
    logger.info(f'df_scores examples: {df_scores.shape}:')

    save_scores(df_scores, conf.inference.output)


if __name__ == '__main__':
    main()



import pickle

import torch

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.trx_encoder import TrxEncoder


def test_example():
    B, T = 5, 20
    trx_encoder = TrxEncoder(
        embeddings={'mcc_code': {'in': 100, 'out': 5}},
        numeric_values={'amount': 'log'},
    )
    x = PaddedBatch(
        payload={
            'mcc_code': torch.randint(0, 99, (B, T)),
            'amount': torch.randn(B, T),
        },
        length=torch.randint(10, 20, (B,)),
    )
    z = trx_encoder(x)
    assert z.payload.shape == (5, 20, 6)  # B, T, H
    assert trx_encoder.output_size == 6


def test_pickle():
    params = {
        "norm_embeddings": False,
        "embeddings_noise": .1,
        'embeddings': {
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        'numeric_values': {'amount': 'log'}
    }

    te = TrxEncoder(**params)
    out = pickle.dumps(te)
    assert len(out) > 0


def test_linear_projection():
    te = TrxEncoder(
        embeddings={
            'mcc_code': {'in': 21, 'out': 2},
            'trans_type': {'in': 11, 'out': 2},
        },
        numeric_values={'amount': 'log'},
        linear_projection_size=11,
    )
    out = te(PaddedBatch({
        'mcc_code': torch.randint(0, 10, (4, 8)),
        'trans_type': torch.randint(0, 10, (4, 8)),
        'amount': torch.randn(4, 8),
    }, torch.LongTensor([10, 4, 9, 2])))
    assert out.payload.size() == (4, 8, 11)
    assert te.output_size == 11


def test_orthogonal_init():
    te = TrxEncoder(
        embeddings={
            'mcc_code': {'in': 5, 'out': 4},
        },
        orthogonal_init=True,
    )
    out = te(PaddedBatch({
        'mcc_code': torch.arange(5).unsqueeze(0),
    }, torch.LongTensor([5]))).payload[0]

    torch.testing.assert_close(
        torch.einsum('ij,kj->ik', out[1:], out[1:]),
        torch.eye(4),
    )


def test_orthogonal_init_with_linear_head():
    te = TrxEncoder(
        embeddings={
            'mcc_code': {'in': 5, 'out': 4},
        },
        linear_projection_size=4,
        orthogonal_init=True,
    )
    out = te.linear_projection_head.weight.data
    torch.testing.assert_close(
        torch.einsum('ij,kj->ik', out, out),
        torch.eye(4),
    )


def test_batch_norm_none():
    te = TrxEncoder(
        numeric_values={
            'amount': 'identity',
        },
        use_batch_norm=False,
    )
    x = PaddedBatch({
        'amount': torch.FloatTensor([
            [1, 2, -1, 3, 0, 0],
            [-2, 1, 0, 0, 0, 0],
        ]),
    }, torch.LongTensor([4, 2]))
    out = te(x).payload

    assert out[0, 0] == 1


def test_batch_norm_batch():
    te = TrxEncoder(
        numeric_values={
            'amount': 'identity',
        },
        use_batch_norm=True,
        use_batch_norm_with_lens=False,
    )
    x = PaddedBatch({
        'amount': torch.FloatTensor([
            [1, 2, -1, 3, 0, 0],
            [-2, 1, 0, 0, 0, 0],
        ]),
    }, torch.LongTensor([4, 2]))
    out = te(x).payload

    flat = torch.FloatTensor([1, 2, -1, 3, 0, 0, -2, 1, 0, 0, 0, 0])
    mean = flat.mean()
    std = flat.std()
    assert abs(out[0, 0] - (1 - mean) / std) < 0.1


def test_batch_norm_with_lens():
    te = TrxEncoder(
        numeric_values={
            'amount': 'identity',
        },
        use_batch_norm=True,
        use_batch_norm_with_lens=True,
    )
    x = PaddedBatch({
        'amount': torch.FloatTensor([
            [1, 2, -1, 3, 0, 0],
            [-2, 1, 0, 0, 0, 0],
        ]),
    }, torch.LongTensor([4, 2]))
    out = te(x).payload

    flat = torch.FloatTensor([1, 2, -1, 3, -2, 1])
    mean = flat.mean()
    std = flat.std()
    assert abs(out[0, 0] - (1 - mean) / std) < 0.1





import numpy as np
import torch
from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class FilterNonArray(IterableProcessingDataset):
    """Deprecated. Use FeatureFilter.

    """
    def __init__(self):
        super().__init__()

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            to_del = []
            for k, v in features.items():
                if isinstance(v, np.ndarray) or isinstance(v, torch.Tensor):
                    continue
                else:
                    to_del.append(k)

            for k in to_del:
                del features[k]
            yield rec



import numpy as np
from ptls.data_load.feature_dict import FeatureDict


class DropoutTrx(FeatureDict):
    """
    This class is used as 'f_augmentation' argument for 
    ptls.data_load.datasets.augmentation_dataset.AugmentationDataset (AugmentationIterableDataset).
    """
    def __init__(self, trx_dropout):
        self.trx_dropout = trx_dropout

    def __call__(self, x):
        seq_len = FeatureDict.get_seq_len(x)

        idx = self.get_idx(seq_len)
        new_x = self.seq_indexing(x, idx)
        return new_x

    def get_idx(self, seq_len):
        if self.trx_dropout > 0 and seq_len > 0:
            idx = np.random.choice(seq_len, size=int(seq_len * (1 - self.trx_dropout)+1), replace=False)
            return np.sort(idx)
        else:
            return np.arange(seq_len)



# coding: utf-8
"""

"""
import functools
import operator

from torch.utils.data import DataLoader

from ptls.data_load import padded_collate, ConvertingTrxDataset, DropoutTrxDataset, ProcessDataset
from ptls.metric_learn.dataset.splitting_dataset import SplittingDataset, SeveralSplittingsDataset
from ptls.frames.coles import split_strategy
from ptls.metric_learn.dataset.target_enumerator_dataset import TargetEnumeratorDataset
from ptls.metric_learn.dataset.complex_target_dataset import ComplexTargetDataset


def nested_list_to_flat_with_collate(collate_fn):
    """Deprecated
    """
    def _func(batch):
        batch = functools.reduce(operator.iadd, batch)
        return collate_fn(batch)

    return _func


def collate_splitted_rows(batch):
    # add Y and flatten samples
    batch = functools.reduce(operator.iadd, batch)
    return padded_collate(batch)


def create_train_data_loader(data, conf):
    dataset = SplittingDataset(
        data,
        split_strategy.create(**conf.params.train.split_strategy)
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    dataset = DropoutTrxDataset(
        dataset,
        trx_dropout=conf.params.train.trx_dropout,
        seq_len=conf.params.train.max_seq_len)

    data_loader = DataLoader(
        dataset,
        shuffle=True,
        collate_fn=collate_splitted_rows,
        num_workers=conf.params.train.num_workers
    )
    return data_loader


def create_valid_data_loader(data, conf, sample_process_fun=None):
    dataset = SplittingDataset(
        data,
        split_strategy.create(**conf.params.valid.split_strategy)
    )
    dataset = TargetEnumeratorDataset(dataset)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=0.0, seq_len=conf.params.valid.max_seq_len)

    data_loader = DataLoader(
        dataset,
        batch_size=conf.params.valid.batch_size,
        shuffle=False,
        collate_fn=collate_splitted_rows,
        num_workers=conf.params.valid.num_workers,
    )
    return data_loader


def create_valid_complex_data_loader(data, conf, sample_process_fun=None):
    split_strategies = [
        split_strategy.create(**split_strategy_params) for split_strategy_params in
        conf.params.valid.split_strategies
    ]
    split_counts = [
        split_strategy_params.split_count for split_strategy_params in conf.params.valid.split_strategies
    ]
    dataset = SeveralSplittingsDataset(
        data,
        split_strategies
    )
    dataset = ComplexTargetDataset(dataset, split_counts)
    dataset = ConvertingTrxDataset(dataset)
    if sample_process_fun is not None:
        dataset = ProcessDataset(dataset, sample_process_fun)
    dataset = DropoutTrxDataset(dataset, trx_dropout=0.0, seq_len=conf.params.valid.max_seq_len)

    data_loader = DataLoader(
        dataset,
        batch_size=conf.params.valid.batch_size,
        shuffle=False,
        collate_fn=collate_splitted_rows,
        num_workers=conf.params.valid.num_workers,
    )
    return data_loader



import torch

from ptls.frames.coles.sampling_strategies import AllPositivePairSelector, HardNegativePairSelector, \
    DistanceWeightedPairSelector


def get_data():
    b, c, h = 3, 2, 2  # Batch, num Classes, Hidden size
    x = torch.tensor([[0., 1.],
                      [1., 0.],
                      [1., 1.],
                      [0.1, 0.9],
                      [0.8, 0.2],
                      [1.0, 0.9]])

    y = torch.arange(c).view(-1, 1).expand(c, b).reshape(-1)
    return x, y


def check_positive_pairs(positive_pairs, y):
    # positives from same classes
    assert torch.equal(y[positive_pairs[:, 0]], y[positive_pairs[:, 1]])


def check_negative_pairs(negative_pairs, y):
    # positives from different classes
    assert torch.all(~torch.eq(y[negative_pairs[:, 0]], y[negative_pairs[:, 1]]))


def test_all_pair_selector():
    x, y = get_data()
    sampling_strategy = AllPositivePairSelector(balance=False)

    positive_pairs, negative_pairs = sampling_strategy.get_pairs(x, y)
    check_positive_pairs(positive_pairs, y)
    check_negative_pairs(negative_pairs, y)

    true_positive_pairs = torch.LongTensor([[0, 1],
                                            [0, 2],
                                            [1, 2],
                                            [3, 4],
                                            [3, 5],
                                            [4, 5]])

    true_negative_pairs = torch.LongTensor([[0, 3],
                                            [0, 4],
                                            [0, 5],
                                            [1, 3],
                                            [1, 4],
                                            [1, 5],
                                            [2, 3],
                                            [2, 4],
                                            [2, 5]])

    assert torch.equal(positive_pairs, true_positive_pairs)
    assert torch.equal(negative_pairs, true_negative_pairs)


def test_hard_pair_selector():
    x, y = get_data()
    sampling_strategy = HardNegativePairSelector()

    positive_pairs, negative_pairs = sampling_strategy.get_pairs(x, y)
    check_positive_pairs(positive_pairs, y)
    check_negative_pairs(negative_pairs, y)

    true_positive_pairs = torch.LongTensor([[0, 1],
                                            [0, 2],
                                            [1, 2],
                                            [3, 4],
                                            [3, 5],
                                            [4, 5]])

    true_negative_pairs = torch.LongTensor([[0, 3],
                                            [1, 4],
                                            [2, 5],
                                            [3, 0],
                                            [4, 1],
                                            [5, 2]])

    assert torch.equal(positive_pairs, true_positive_pairs)
    assert torch.equal(negative_pairs, true_negative_pairs)


def test_distance_weighted_pair_selector():
    x, y = get_data()
    sampling_strategy = DistanceWeightedPairSelector(batch_k=3)
    positive_pairs, negative_pairs = sampling_strategy.get_pairs(x, y)
    check_positive_pairs(positive_pairs, y)
    check_negative_pairs(negative_pairs, y)



import torch


class AbsSeqEncoder(torch.nn.Module):
    def __init__(self, is_reduce_sequence=True):
        super().__init__()
        self._is_reduce_sequence = is_reduce_sequence

    @property
    def is_reduce_sequence(self):
        return self._is_reduce_sequence

    @is_reduce_sequence.setter
    def is_reduce_sequence(self, value):
        self._is_reduce_sequence = value

    @property
    def embedding_size(self):
        raise NotImplementedError()



from typing import List

import torch
from torch.nn import Linear, BatchNorm1d, Sigmoid, Sequential, ReLU, LogSoftmax, Flatten, Softplus, Dropout
from ptls.nn.normalization import L2NormEncoder


class Head(torch.nn.Module):
    r"""Head for the sequence encoder

    Parameters
    ----------
         input_size: int
            input size
         use_norm_encoder: bool. Default: False
            whether to use normalization layers before the head
         use_batch_norm: bool. Default: False.
            whether to use BatchNorm.
         hidden_layers_sizes: List[int]. Default: None.
            sizes of linear layers. If None without additional linear layers.
         objective: str. Options:
            None (default) - corresponds to linear output with relu
            classification - linear output with sigmoid or logsoftmax (num_classes > 1)
            regression - pure linear output
            softplus - linear output with softplus
         num_classes: int. Default: 1.
            The number of classed in classification problem. Default correspond to binary classification.

     """
    def __init__(self,
                 input_size: int = None,
                 use_norm_encoder: bool = False,
                 use_batch_norm: bool = False,
                 hidden_layers_sizes: List[int] = None,
                 drop_probs: List[float] = None,
                 objective: str = None,
                 num_classes: int = 1):
        super().__init__()
        # TODO: check possibility to create empty head with do nothing

        layers = []

        if use_norm_encoder:
            layers.append(L2NormEncoder())

        if use_batch_norm:
            layers.append(BatchNorm1d(input_size))

        if drop_probs: assert len(drop_probs) == len(hidden_layers_sizes),\
            'dimensions of `drop_probs` and `hidden_layers_sizes` should be equal'

        if hidden_layers_sizes is not None:
            layers_size = [input_size] + list(hidden_layers_sizes)
            for ix, (size_in, size_out) in enumerate(zip(layers_size[:-1], layers_size[1:])):
                layers.append(Linear(size_in, size_out))
                if use_batch_norm:
                    layers.append(BatchNorm1d(size_out))
                layers.append(ReLU())
                if drop_probs:
                    layers.append(Dropout(drop_probs[ix]))
                input_size = size_out

        if objective == 'classification':
            if num_classes == 1:
                h = Sequential(Linear(input_size, num_classes), Sigmoid(), Flatten(0))
            else:
                h = Sequential(Linear(input_size, num_classes), LogSoftmax(dim=1))
            layers.append(h)

        elif objective == 'regression':
            if num_classes == 1:
                layers.append(Sequential(Linear(input_size, 1), Flatten(0)))
            else:
                layers.append(Linear(input_size, num_classes))

        elif objective == 'softplus':
            if num_classes == 1:
                layers.append(Sequential(Linear(input_size, num_classes), Softplus(), Flatten(0)))
            else:
                layers.append(Sequential(Linear(input_size, num_classes), Softplus()))

        elif objective is not None:
            raise AttributeError(f"Unknown objective {objective}. Supported: classification, regression and softplus.")

        self.model = torch.nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)



import os
import warnings
from glob import glob
from typing import Union, List
from omegaconf import ListConfig
from sklearn.model_selection import train_test_split


def _scan_path(path):
    if os.path.isfile(path):
        return [path]
    if os.path.isdir(path):
        file_path = os.path.join(path, '*.parquet')
        return glob(file_path)
    return []


def train_valid_split(
        data,
        valid_rate: Union[float, int] = None,
        is_sorted: bool =True,
        return_part: str = 'train',
        shuffle_seed: int = 42,
):
    """

    Parameters
    ----------
    data
        objects for split
    valid_rate:
        if set split found files into train-test
        int means valid objects count, float means valid objects rate
    is_sorted:
        sort or not found files. Should be True when `valid_rate` split used
    return_part: one of ['train', 'valid']
        Which part will be returned when `valid_rate` split used
    shuffle_seed:
        random seed for train_test_split
    Returns
    -------
        object list which are the same as `data` when `valid_rate` aren't used or part of `data` if `valid_rate` used
    """
    if is_sorted:
        data = sorted(data)

    if valid_rate is None or valid_rate == 0.0 or valid_rate == 0:
        return data

    if valid_rate is not None and not is_sorted:
        warnings.warn('train_test_split on unsorted data may returns unexpected result. '
                      'Use `is_sorted=True` when `valid_rate > 0.0`')

    train, valid = train_test_split(data, test_size=valid_rate, random_state=shuffle_seed)
    if return_part == 'train':
        return train
    elif return_part == 'valid':
        return valid
    else:
        raise AttributeError('Never happens')


def parquet_file_scan(
        file_path: Union[str, List[str], ListConfig],
        valid_rate: Union[float, int] = None,
        is_sorted: bool =True,
        return_part: str = 'train',
        shuffle_seed: int = 42,
):
    """Scan folder with parquet files and returns file names. Train-valid split possible

    Split should be reproducible with same results when `is_sorted=True` and other parameters don't change.
    This means that you can split files into synchronised train-valid parts with two calls.

    Example:
    >>> files = [1, 2, 3, 4, 6, 5, 9, 8]
    >>> train_files = train_valid_split(files, valid_rate=0.3, return_part='train', shuffle_seed=123)
    >>> valid_files = train_valid_split(files, valid_rate=0.3, return_part='valid', shuffle_seed=123)
    >>> for i in train_files:
    >>>     assert i not in valid_files
    >>> for i in valid_files:
    >>>     assert i not in train_files

    Parameters
    ----------
    file_path:
        path for scan. Can be single file, directory or list of them.
    valid_rate:
        if set split found files into train-test
        int means valid objects count, float means valid objects rate
    is_sorted:
        sort or not found files. Should be True when `valid_rate` split used
    return_part: one of ['train', 'valid']
        Which part will be returned when `valid_rate` split used
    shuffle_seed:
        random seed for train_test_split
    Returns
    -------
        File list which are all found files when `valid_rate` aren't used or part of files if `valid_rate` used

    """
    assert return_part in ('train', 'valid')

    if type(file_path) not in (list, ListConfig):
        file_path = [file_path]

    data_files = []
    for path in file_path:
        data_files.extend(_scan_path(path))

    return train_valid_split(data_files, valid_rate, is_sorted, return_part, shuffle_seed)



import os
from glob import glob

import pandas as pd
import yaml
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator
from tqdm import tqdm


def get_versions(path):
    version_list = glob(os.path.join(path, 'version_*'))
    for version_path in version_list:
        yield version_path, os.path.split(version_path)[-1]


def get_scalars(path):
    df_list = []
    for version_path, version_name in get_versions(path):
        events_path = glob(os.path.join(version_path, 'events.*'))[0]
        event_acc = EventAccumulator(events_path)
        event_acc.Reload()

        df_list.append(pd.DataFrame((
            (version_name, tag, e.step, e.value)
            for tag in event_acc.Tags()['scalars']
            for e in event_acc.Scalars(tag)
        ), columns=['version', 'tag', 'step', 'value']))
    return pd.concat(df_list, axis=0)


def get_last_scalars(path):
    df = get_scalars(path)
    df = df[lambda x: x['step'].gt(0)]
    df = df.set_index(['version', 'tag', 'step']).sort_index()
    df = df.groupby(['version', 'tag'])['value'].last().unstack()

    return df


def get_hparams(path, progress=False):
    df_list = []
    for version_path, version_name in tqdm(get_versions(path), disable=not progress):
        hp_path = glob(os.path.join(version_path, 'hparams.yaml'))[0]
        with open(hp_path, 'r') as f:
            hp = yaml.load(f, Loader=yaml.UnsafeLoader)
            hp = pd.DataFrame([hp], index=[version_name])
            df_list.append(hp)
    return pd.concat(df_list, axis=0)



import torch

from ptls.data_load import PaddedBatch
from ptls.nn.trx_encoder.batch_norm import RBatchNorm, RBatchNormWithLens


def test_r_batch_norm():
    m = RBatchNorm(1)
    x = PaddedBatch(
        torch.randn(20, 400, 1),
        torch.arange(20, dtype=torch.long) + 10,
    )
    out = m(x).payload
    assert out.size() == (20, 400, 1)


def test_r_batch_norm_with_lens():
    m = RBatchNormWithLens(1)
    x = PaddedBatch(
        torch.randn(20, 400, 1),
        torch.arange(20, dtype=torch.long) + 10,
    )
    out = m(x).payload
    assert out.size() == (20, 400, 1)



import numpy as np


class ListSplitter:
    """
    Collect list of objects and split it on train and valid (if possible)
    Returns list of objects for train and valid
    """
    def __init__(self, object_list, valid_size=0.5, is_sort=True, seed=None):
        self.object_list = object_list
        self.valid_size = valid_size
        self.is_sort = is_sort
        self.seed = seed

        self._n = len(self.object_list)
        self._train_files = None
        self._valid_files = None

        self.split()
        self.sort()

    @property
    def train(self):
        return self._train_files

    @property
    def valid(self):
        return self._valid_files

    def size_select(self):
        if self.valid_size == 0.0:
            return self._n, 0

        _valid = self._n * self.valid_size
        if _valid < 1.0:
            _valid = 1
        else:
            _valid = int(round(_valid))

        _train = self._n - _valid
        if _train == 0:
            raise AttributeError(', '.join([
                'Incorrect train size',
                f'N: {self._n}',
                f'train: {_train}',
                f'valid: {_valid}',
                f'rate: {self.valid_size:.3f}',
            ]))
        return _train, _valid

    def split(self):
        train_size, valid_size = self.size_select()

        if valid_size == 0:
            self._train_files = self.object_list
            self._valid_files = None
        else:
            valid_ix = np.arange(self._n)
            rs = np.random.RandomState(self.seed)
            valid_ix = set(rs.choice(valid_ix, size=valid_size, replace=False).tolist())

            self._train_files = [rec for i, rec in enumerate(self.object_list) if i not in valid_ix]
            self._valid_files = [rec for i, rec in enumerate(self.object_list) if i in valid_ix]

    def sort(self):
        if not self.is_sort:
            return
        if self._train_files is not None:
            self._train_files = sorted(self._train_files)
        if self._valid_files is not None:
            self._valid_files = sorted(self._valid_files)
